# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest

# Allow everything in there to access the DB
pytestmark = pytest.mark.django_db

import copy
from django.core.urlresolvers import reverse
import json
import logging
from rest_framework import status

from .fixtures import live_server, client, user, site
from .util import (
    assert_created, assert_error, assert_success, assert_deleted, load_json,
    Client, load
)


log = logging.getLogger(__name__)


def test_permissions(live_server, user, site):
    admin_client = Client(live_server, 'admin')
    user_client = Client(live_server, 'user')

    # URIs
    attr_uri = site.list_uri('attribute')
    net_uri = site.list_uri('network')
    dev_uri = site.list_uri('device')

    # Create an Attribute
    attr_resp = admin_client.create(
        attr_uri, resource_name='Network', name='attr1'
    )
    attr = attr_resp.json()['data']['attribute']
    attr_obj_uri = site.detail_uri('attribute', id=attr['id'])

    # Create a Network
    net_resp = admin_client.create(net_uri, cidr='10.0.0.0/24')
    net = net_resp.json()['data']['network']
    net_obj_uri = site.detail_uri('network', id=net['id'])

    # Create a Device
    dev_resp = admin_client.create(dev_uri, hostname='dev1')
    dev= dev_resp.json()['data']['device']
    dev_obj_uri = site.detail_uri('device', id=dev['id'])

    # User shouldn't be able to update site or create/update other resources
    # Site
    assert_error(
        user_client.update(site.detail_uri(), name='site1'),
        status.HTTP_403_FORBIDDEN
    )
    # Attribute
    assert_error(
        user_client.create(attr_uri, name='attr2'),
        status.HTTP_403_FORBIDDEN
    )
    assert_error(
        user_client.update(attr_obj_uri, required=True),
        status.HTTP_403_FORBIDDEN
    )
    # Network
    assert_error(
        user_client.create(net_uri, cidr='10.0.0.0/8'),
        status.HTTP_403_FORBIDDEN
    )
    assert_error(
        user_client.update(net_obj_uri, attributes={'attr1': 'foo'}),
        status.HTTP_403_FORBIDDEN
    )
    # Device
    assert_error(
        user_client.create(dev_uri, name='dev2'),
        status.HTTP_403_FORBIDDEN
    )
    assert_error(
        user_client.update(dev_obj_uri, hostname='foobar'),
        status.HTTP_403_FORBIDDEN
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest

# Allow everything in there to access the DB
pytestmark = pytest.mark.django_db

import copy
from django.core.urlresolvers import reverse
import json
import logging
from rest_framework import status


from .fixtures import live_server, client, user, site
from .util import (
    assert_created, assert_error, assert_success, assert_deleted, load_json,
    Client, load
)


log = logging.getLogger(__name__)


def test_user_with_secret_key(live_server):
    user1_client = Client(live_server, 'user1')
    user2_client = Client(live_server, 'user2')

    # URI for user 0
    user_uri = reverse('user-detail', args=(0,))

    # Small requests to make user accounts in order.
    user1_resp = user1_client.get(user_uri)
    user1 = user1_resp.json()['data']['user']
    user1_uri = reverse('user-detail', args=(user1['id'],))

    user2_resp = user2_client.get(user_uri)
    user2 = user2_resp.json()['data']['user']
    user2_uri = reverse('user-detail', args=(user2['id'],))

    # User should be able to get user 0 (self)
    assert_success(
        user1_client.get(user_uri),
        {'user': user1}
    )

    # And see their own secret key as user 0
    response = user1_client.get(user_uri + '?with_secret_key')
    expected = copy.deepcopy(user1)
    expected['secret_key'] = response.json()['data']['user']['secret_key']
    assert_success(response, {'user': expected})

    # And their own secret key by their user id 
    response = user1_client.get(user1_uri + '?with_secret_key')
    assert_success(response, {'user': expected})

    # But not user 2's secret_key.
    response = user1_client.get(user2_uri + '?with_secret_key')
    assert_error(response, status.HTTP_403_FORBIDDEN)


def test_user_rotate_secret_key(live_server):
    user1_client = Client(live_server, 'user1')
    user2_client = Client(live_server, 'user2')

    # URI for user 0
    user_uri = reverse('user-detail', args=(0,))

    # Small requests to make user accounts in order.
    user1_resp = user1_client.get(user_uri)
    user1 = user1_resp.json()['data']['user']
    user1_key_uri = reverse('user-rotate-secret-key', args=(user1['id'],))

    user2_resp = user2_client.get(user_uri)
    user2 = user2_resp.json()['data']['user']
    user2_key_uri = reverse('user-rotate-secret-key', args=(user2['id'],))

    # User1 should be able to rotate their own secret_key
    assert_success(user1_client.post(user1_key_uri))

    # But not user 2's secret_key
    assert_error(user1_client.post(user2_key_uri), status.HTTP_403_FORBIDDEN)



from __future__ import absolute_import, unicode_literals

from django.conf.urls import include, url
from django.contrib import admin

from ..api.views import NotFoundViewSet
from ..ui.views import FeView


# Custom error-handling views.
handler400 = 'nsot.ui.views.handle400'
handler403 = 'nsot.ui.views.handle403'
handler404 = 'nsot.ui.views.handle404'
handler500 = 'nsot.ui.views.handle500'


urlpatterns = [
    # API
    url(r'^api/', include('nsot.api.urls')),

    # Catchall for missing endpoints
    url(r'^api/.*/$', NotFoundViewSet.as_view({'get': 'list'})),

    # Docs (Swagger)
    url(r'^docs/', include('rest_framework_swagger.urls')),

    # Admin
    url(r'^admin/', include(admin.site.urls)),

    # FE handlers
    # Catch index
    url(r'^$', FeView.as_view(), name='index'),

    # Catch all for remaining URLs
    url(r'^.*/$', FeView.as_view(), name='index'),
]



from __future__ import unicode_literals

from collections import namedtuple, OrderedDict
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.core.exceptions import ValidationError as DjangoValidationError
from django.db.models import Q, ProtectedError
from django.shortcuts import render
import logging
from rest_framework import mixins, status, viewsets
from rest_framework.views import APIView
from rest_framework.decorators import detail_route, list_route
from rest_framework.response import Response
from rest_framework_bulk import mixins as bulk_mixins

from . import auth
from . import serializers
from .. import exc
from .. import models
from ..util import qpbool, parse_set_query


log = logging.getLogger(__name__)


class BaseNsotViewSet(viewsets.ReadOnlyModelViewSet):
    """
    Default viewset for Nsot objects with the following defaults:

    + Successful responses are in format ``{"status": "ok", "data": DATA}``
    + Error responses are in format ``{"status": "error", "error": ERROR}``
    + All list results always display pagination controls
    + Objects are designed to be nested under site resources, but can also be
      top-level resources.
    """
    def __init__(self, *args, **kwargs):
        super(BaseNsotViewSet, self).__init__(*args, **kwargs)

        # This is the model's human-readable name for results and error
        # messages.
        self.result_key = self.queryset.model._meta.model_name
        self.result_key_plural = self.result_key + 's'
        self.model_name = self.result_key.title()

    def not_found(self, pk=None, site_pk=None, msg=None):
        """Standard formatting for 404 errors."""
        if msg is None:
            msg = 'No such {} found at (site_id, id) = ({}, {})'.format(
                    self.model_name, site_pk, pk
                )
        raise exc.NotFound(msg)

    def success(self, data, result_key=None, status=200, headers=None):
        if result_key is None:
            result_key = self.result_key

        # If there are multiple objects, use the plural result_key
        if isinstance(data, list):
            result_key = self.result_key_plural

        return Response(
            OrderedDict([
                ('status', 'ok'),
                ('data', {result_key: data}),
            ]),
            status=status,
            headers=headers,
        )

    def get_paginated_response(self, data, result_key=None):
        """Overload default pagination to customize `result_key`."""
        assert self.paginator is not None
        return self.paginator.get_paginated_response(data, result_key)

    def create(self, request, *args, **kwargs):
        """Return objects that have just been created."""
        response = super(BaseNsotViewSet, self).create(request, *args, **kwargs)
        return self.success(
            response.data, status=response.status_code, headers=dict(response.items()),
        )

    def update(self, request, *args, **kwargs):
        """Return objects that have just been updated."""
        response = super(BaseNsotViewSet, self).update(request, *args, **kwargs)
        return self.success(response.data)

    def list(self, request, site_pk=None, queryset=None, *args, **kwargs):
        """List objects optionally filtered by site."""
        if queryset is None:
            queryset = self.filter_queryset(self.get_queryset())
            # Query by site_pk if it's set (.e.g /site/1/:foo) and make sure any
            # filtering args are passed.
            if site_pk is not None:
                queryset = queryset.filter(site=site_pk)

        page = self.paginate_queryset(queryset)
        serializer = self.get_serializer(page, many=True)
        return self.get_paginated_response(
            data=serializer.data, result_key=self.result_key_plural
        )

    def retrieve(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Retrieve a single object optionally filtered by site."""
        try:
            if site_pk is not None:
                obj = self.queryset.get(pk=pk, site=site_pk)
            else:
                obj = self.queryset.get(pk=pk)
        except ObjectDoesNotExist:
            self.not_found(pk, site_pk)

        serializer = self.get_serializer(obj, *args, **kwargs)
        return self.success(serializer.data)


class ChangeViewSet(BaseNsotViewSet):
    """
    Read-only API endpoint that allows Changes to be viewed.

    All Create/Update/Delete events are logged as a Change. A Change includes
    information such as the change time, user, and the full resource after
    modification. Changes are immutable and can only be removed by deleting the
    entire Site.
    """
    queryset = models.Change.objects.order_by('-change_at')
    serializer_class = serializers.ChangeSerializer
    filter_fields = ('event', 'resource_name', 'resource_id')


class NsotViewSet(BaseNsotViewSet, viewsets.ModelViewSet):
    """
    Generic mutable viewset that logs all change events and includes support
    for bulk creation of objects.
    """
    def perform_create(self, serializer):
        """Support bulk create."""
        try:
            objects = serializer.save()
        except DjangoValidationError as err:
            raise exc.ValidationError(err.error_dict)
        else:
            # This is so that we can always work w/ objects as a list
            if not isinstance(objects, list):
                objects = [objects]

        log.debug('NsotViewSet.perform_create() objects = %r', objects)
        for obj in objects:
            models.Change.objects.create(
                obj=obj, user=self.request.user, event='Create'
            )

    def get_success_headers(self, data):
        # TODO(jathan): Implement hyperlinked fields in the API?
        try:
            # return {'Location': data[api_settings.URL_FIELD_NAME]}
            location = '%s%s/' % (self.request.path_info, data['id'])
            return {'Location': location}
        except (TypeError, KeyError):
            return {}

    def perform_update(self, serializer):
        try:
            obj = serializer.save()
        except DjangoValidationError as err:
            raise exc.ValidationError(err.error_dict)

        log.debug('NsotViewSet.perform_update() obj = %r', obj)
        models.Change.objects.create(
            obj=obj, user=self.request.user, event='Update'
        )

    def perform_destroy(self, instance):
        log.debug('NsotViewSet.perform_destroy() obj = %r', instance)
        models.Change.objects.create(
            obj=instance, user=self.request.user, event='Delete'
        )

        try:
            instance.delete()
        except ProtectedError as err:
            raise exc.Conflict(err.args[0])


class AttributeViewSet(NsotViewSet, bulk_mixins.BulkCreateModelMixin):
    """
    API endpoint that allows Attributes to be viewed or edited.
    """
    queryset = models.Attribute.objects.all()
    serializer_class = serializers.AttributeSerializer
    filter_fields = ('name', 'resource_name')

    def get_serializer_class(self):
        if self.request.method == 'POST':
            return serializers.AttributeCreateSerializer
        if self.request.method == 'PUT':
            return serializers.AttributeUpdateSerializer
        return self.serializer_class

    def get_queryset(self):
        """
        Filter the queryset based on query arguments.
        """
        # Only do this advanced query filtering when we're listing items.
        if self.request.method != 'GET':
            return super(AttributeViewSet, self).get_queryset()

        attributes = self.queryset

        params = self.request.query_params
        required = params.get('required', None)
        display = params.get('display', None)
        multi = params.get('multi', None)

        if required is not None:
            attributes = attributes.filter(required=qpbool(required))
        if display is not None:
            attributes = attributes.filter(display=qpbool(display))
        if multi is not None:
            attributes = attributes.filter(multi=qpbool(multi))

        return attributes


class ValueViewSet(NsotViewSet):
    """
    API endpoint that allows Attribute Values to be viewed or edited.
    """
    queryset = models.Value.objects.all()
    serializer_class = serializers.ValueSerializer

    def get_serializer_class(self):
        if self.request.method == 'POST':
            return serializers.ValueCreateSerializer
        return self.serializer_class


class ResourceViewSet(NsotViewSet, bulk_mixins.BulkCreateModelMixin):
    """
    Resource views that include set query list endpoints.
    """
    @list_route(methods=['get'])
    def query(self, request, site_pk=None, *args, **kwargs):
        """Perform a set query."""
        query = request.query_params.get('query', '')

        objects = self.queryset.set_query(query, site_pk=site_pk)
        return self.list(request, queryset=objects, *args, **kwargs)


class DeviceViewSet(ResourceViewSet):
    """
    API endpoint that allows Devices to be viewed or edited.
    """
    queryset = models.Device.objects.all()
    serializer_class = serializers.DeviceSerializer
    filter_fields = ('hostname',)

    def get_serializer_class(self):
        if self.request.method == 'POST':
            return serializers.DeviceCreateSerializer
        if self.request.method == 'PUT':
            return serializers.DeviceUpdateSerializer

        return self.serializer_class

    def get_queryset(self):
        """This is so we can filter by attribute/value pairs."""
        # Only do this advanced query filtering when we're listing items.
        if self.request.method != 'GET':
            return super(DeviceViewSet, self).get_queryset()

        devices = self.queryset

        params = self.request.query_params
        attributes = params.getlist('attributes', [])

        # Iterate the attributes and try to look them up as if they are k=v
        # and naively do an intersection query.
        log.debug('GOT ATTRIBUTES: %r', attributes)
        for attribute in attributes:
            name, _, value = attribute.partition('=')
            # Retrieve next set of networks using the same arguments as the
            # initial query.
            next_set = Q(
                attributes=models.Value.objects.filter(
                    attribute__name=name, value=value
                )
            )
            devices = devices.filter(next_set)

        return devices


class NetworkViewSet(ResourceViewSet):
    """
    API endpoint that allows Networks to be viewed or edited.
    """
    queryset = models.Network.objects.all()
    serializer_class = serializers.NetworkSerializer

    def get_serializer_class(self):
        if self.request.method == 'POST':
            return serializers.NetworkCreateSerializer
        if self.request.method == 'PUT':
            return serializers.NetworkUpdateSerializer

        return self.serializer_class

    @list_route(methods=['get'])
    def query(self, request, site_pk=None, *args, **kwargs):
        """Override base query to inherit filtering by query params."""
        self.queryset = self.get_queryset()
        return super(NetworkViewSet, self).query(
            request, site_pk, *args, **kwargs
        )

    def get_queryset(self):
        """
        Filter the queryset based on query arguments.
        """
        # Only do this advanced query filtering when we're listing items.
        if self.request.method != 'GET':
            return super(NetworkViewSet, self).get_queryset()

        networks = self.queryset

        params = self.request.query_params
        include_networks = qpbool(params.get('include_networks', True))
        include_ips = qpbool(params.get('include_ips', False))
        root_only = qpbool(params.get('root_only', False))
        cidr = params.get('cidr', None)
        attributes = params.getlist('attributes', [])
        network_address = params.get('network_address', None)
        prefix_length = params.get('prefix_length', None)

        if not any([include_networks, include_ips]):
            return networks.none()

        if not all([include_networks, include_ips]):
            if include_networks:
                networks = networks.filter(is_ip=False)
            if include_ips:
                networks = networks.filter(is_ip=True)

        if root_only:
            networks = networks.filter(parent=None)

        # If cidr is provided, use it to populate network_address and prefix_length
        if cidr is not None:
            log.debug('got cidr: %s' % cidr)
            network_address, _, prefix_length = cidr.partition('/')

        # If network_address is provided, pack it.
        if network_address is not None:
            log.debug('got network_address: %s' % network_address)
            networks = networks.filter(network_address=network_address)

        # If prefix_length is provided, convert it to an int.
        if prefix_length is not None:
            log.debug('got prefix_length: %s' % prefix_length)
            try:
                prefix_length = int(prefix_length)
            except ValueError:
                raise exc.BadRequest(
                    'Invalid prefix_length: %s' % prefix_length
                )
            networks = networks.filter(prefix_length=prefix_length)

        # Iterate the attributes and try to look them up as if they are k=v
        # and naively do an intersection query.
        log.debug('GOT ATTRIBUTES: %r', attributes)
        for attribute in attributes:
            name, _, value = attribute.partition('=')
            # Retrieve next set of networks using the same arguments as the
            # initial query.
            next_set = Q(
                attributes=models.Value.objects.filter(
                    attribute__name=name, value=value
                )
            )
            networks = networks.filter(next_set)

        return networks

    @detail_route(methods=['get'])
    def subnets(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return subnets."""
        if site_pk is not None:
            query = self.queryset.filter(pk=pk, site=site_pk)
        else:
            query = self.queryset.filter(pk=pk)
        network = query.first()

        if not network:
            self.not_found(pk, site_pk)

        params = request.query_params
        include_networks = qpbool(params.get('include_networks', True))
        include_ips = qpbool(params.get('include_ips', False))
        direct = qpbool(params.get('direct', False))

        networks = network.subnets(
            include_networks=include_networks,
            include_ips=include_ips, direct=direct
        )

        return self.list(request, queryset=networks, *args, **kwargs)

    @detail_route(methods=['get'])
    def supernets(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Return supernets."""
        if site_pk is not None:
            query = self.queryset.filter(pk=pk, site=site_pk)
        else:
            query = self.queryset.filter(pk=pk)
        network = query.first()

        if not network:
            self.not_found(pk, site_pk)

        params = request.query_params
        direct = qpbool(params.get('direct', False))

        networks = network.supernets(direct=direct)

        return self.list(request, queryset=networks, *args, **kwargs)


class SiteViewSet(NsotViewSet):
    """
    API endpoint that allows Sites to be viewed or edited.
    """
    queryset = models.Site.objects.all()
    serializer_class = serializers.SiteSerializer
    filter_fields = ('name',)


#: Namedtuple for retrieving pk and user object of current user.
UserPkInfo = namedtuple('UserPkInfo', 'user pk')


class UserViewSet(BaseNsotViewSet, mixins.CreateModelMixin):
    """
    This viewset automatically provides `list` and `detail` actins.
    """
    queryset = get_user_model().objects.all()
    serializer_class = serializers.UserSerializer
    filter_fields = ('email',)

    def get_user_and_pk(self, request, pk=None, site_pk=None):
        # If pk is 0, return the current user.
        if int(pk) == 0:
            pk = request.user.pk  # Authenticated user

        # Try to get the requested user
        try:
            user = self.queryset.get(pk=pk)
        except ObjectDoesNotExist:
            self.not_found(pk, site_pk)

        return UserPkInfo(user, pk)

    def retrieve(self, request, pk=None, site_pk=None, *args, **kwargs):
        """Retreive a single user."""
        user, pk = self.get_user_and_pk(request, pk, site_pk)

        params = request.query_params
        with_secret_key = params.get('with_secret_key', None)

        # If with_secret_key is set, confirm that the requested user object
        # matches the current user.
        if with_secret_key is not None:
            if user != request.user:
                raise exc.Forbidden(
                    "Can't access secret_key of user that isn't you."
                )
            kwargs['with_secret_key'] = qpbool(with_secret_key)

        return super(UserViewSet, self).retrieve(
            request, pk, site_pk, *args, **kwargs
        )

    @detail_route(methods=['post'])
    def rotate_secret_key(self, request, pk=None, *args, **kwargs):
        user, pk = self.get_user_and_pk(request, pk)

        if user != request.user:
            raise exc.Forbidden(
                "Can't access secret_key of user that isn't you."
        )

        user.rotate_secret_key()
        return self.success(user.secret_key, 'secret_key')


class NotFoundViewSet(viewsets.GenericViewSet):
    """Catchall for bad API endpoints."""
    def get(self, *args, **kwargs):
        raise exc.NotFound()

    def list(self, *args, **kwargs):
        raise exc.NotFound()

    def get_serializer_class(self):
        return None


class AuthTokenLoginView(APIView):
    permission_classes = ()

    def post(self, request, *args, **kwargs):
        serializer = serializers.AuthTokenSerializer(data=request.data)

        if serializer.is_valid():
            user = serializer.validated_data['user']
            data = {'auth_token': user.generate_auth_token()}

            return Response(
                OrderedDict([
                    ('status', 'ok'),
                    ('data', data),
                ])
            )
        raise exc.Unauthorized(serializer.errors)


class AuthTokenVerifyView(APIView):
    authentication_classes = (auth.AuthTokenAuthentication,)
    permission_classes = ()

    def post(self, request, *args, **kwargs):
        return Response(
            OrderedDict([
                ('status', 'ok'),
                ('data', True),
            ])
        )



from __future__ import absolute_import, print_function

import sys

import BaseHTTPServer
from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError
import logging
from optparse import make_option


class Command(BaseCommand):
    args = '<username>'
    help = 'Start a development reverse proxy.'

    option_list = BaseCommand.option_list + (
        make_option('--address', '-a',
            dest='address',
            type=str,
            default='localhost',
            help='Address to listen on',
        ),
        make_option('--listen-port', '-p',
            dest='listen_port',
            type=int,
            default=8888,
            help='Port to listen on.'
        ),
        make_option('--backend-port', '-P',
            dest='backend_port',
            type=int,
            default=8990,
            help='Port to proxy to',
        ),
    )

    def handle(self, username, **options):
        try:
            from mrproxy import UserProxyHandler
        except ImportError:
            raise SystemExit(
                'mrproxy is required for the user proxy. Please see README.md.'
            )

        class ServerArgs(object):
            def __init__(self, backend_port, username):
                self.backend_port = backend_port
                self.header = ['X-NSoT-Email: %s' % username]

        if username is None:
            username = getpass.getuser()
            logging.debug('No username provided, using (%s)', username)
        username += '@localhost'


        address = options.get('address')
        listen_port = options.get('listen_port')

        server = BaseHTTPServer.HTTPServer(
            (address, listen_port), UserProxyHandler
        )
        server.args = ServerArgs(options.get('backend_port'), username)
        try:
            logging.info(
                'Starting user_proxy on %s:%s with user %r',
                address, listen_port, username
            )
            server.serve_forever()
        except KeyboardInterrupt:
            print('Bye!')



"""
Make dummy data and fixtures and stuff to use in benchmarking.
"""

import collections
import faker
import json
import hashlib
from hashlib import sha1
import ipaddress
import pytest
import random
import socket
import struct
import time


# Constants and stuff
fake = faker.Factory.create()

# Phony attributes to randomly generate for testing.
ATTRIBUTE_DATA = {
    'lifecycle': ['monitored', 'ignored'],
    'owner': ['jathan', 'gary', 'lisa', 'jimmy', 'bart', 'bob', 'alice'],
    'metro': ['lax', 'iad', 'sjc', 'tyo'],
    'foo': ['bar', 'baz', 'spam'],
}

# Used to store Attribute/value pairs
Attribute = collections.namedtuple('Attribute', 'name value')


def rando():
    return random.choice((True, False))


def generate_words(num_items=100, title=False, add_suffix=False):
    stuff = set()
    for _ in range(num_items + 1):
        things = (fake.word(), fake.first_name(), fake.last_name())
        suffix = str(random.randint(1, 32)) if add_suffix else ''
        word = random.choice(things) + suffix

        # Title it?
        if rando():
            word = word.title()

        # Reverse it?
        if rando():
            word = word[::-1]  # Reverse it

        # Lower it?
        else:
            word = word.lower()

        # stuff.add(word)
        yield word


def generate_hostnames(num_items=100):
    """
    Generate a random list of hostnames.

    :param num_items:
        Number of items to generate
    """

    for i in range(1, num_items + 1):
        yield 'host%s' % i


def generate_ipv4():
    """Generate a random IPv4 address."""
    return socket.inet_ntoa(struct.pack('>I', random.randint(1, 0xffffffff)))


def generate_ipv4list(num_items=100, include_hosts=False):
    """
    Generate a list of unique IPv4 addresses. This is a total hack.

    :param num_items:
        Number of items to generate

    :param include_hosts:
        Whether to include /32 addresses
    """
    ipset = set()
    # Keep iterating and hack together cidr prefixes if we detect empty
    # trailing octects. This is so lame that we'll mostly just end up with a
    # bunch of /24 networks.
    while len(ipset) < num_items:
        ip = generate_ipv4()
        if ip.startswith('0'):
            continue

        if ip.endswith('.0.0.0'):
            prefix = '/8'
        elif ip.endswith('.0.0'):
            prefix = '/16'
        elif ip.endswith('.0'):
            prefix = '/24'
        elif include_hosts:
            prefix = '/32'
        else:
            continue

        ip += prefix
        ipset.add(ip)
    return sorted(ipset)


def enumerate_attributes(resource_name, attributes=None):
    if attributes is None:
        attributes = ATTRIBUTE_DATA

    for name in attributes:
        yield {'name': name, 'resource_name': resource_name}


def generate_attributes(attributes=None, as_dict=True):
    """
    Randomly choose attributes and values for testing.

    :param attributes:
        Dictionary of attribute names and values

    :param as_dict:
        If set return a dict vs. list of Attribute objects
    """
    if attributes is None:
        attributes = ATTRIBUTE_DATA
    attrs = []
    for attr_name, attr_values in attributes.iteritems():
        if random.choice((True, False)):
            attr_value = random.choice(attr_values)
            attrs.append(Attribute(attr_name, attr_value))
    if as_dict:
        attrs = dict(attrs)
    return attrs


def generate_devices(num_items=100, with_attributes=True):
    """
    Return a list of dicts for Device creation.

    :param num_items:
        Number of items to generate

    :param with_attributes:
        Whether to include Attributes
    """
    hostnames = generate_hostnames(num_items)

    devices = []
    for hostname in hostnames:
        item = {'hostname': hostname}
        if with_attributes:
            attributes = generate_attributes()
            item['attributes'] = attributes

        yield item
        # devices.append(item)
    # return devices


def generate_networks(num_items=100, with_attributes=True, include_hosts=False):
    """
    Return a list of dicts for Network creation.

    :param num_items:
        Number of items to generate

    :param with_attributes:
        Whether to include Attributes
    """
    ipv4list = generate_ipv4list(num_items, include_hosts=include_hosts)

    networks = []
    for cidr in ipv4list:
        attributes = generate_attributes()
        item = {'cidr': cidr}
        if with_attributes:
            attributes = generate_attributes()
            item['attributes'] = attributes

        networks.append(item)
    return networks


def rando_set_action():
    return random.choice(['+', '-', ''])


def rando_set_query():
    action = rando_set_action()
    return ' '.join(
        action + '%s=%s' % (k, v) for k,v in generate_attributes().iteritems()
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest
# Allow everything in there to access the DB
pytestmark = pytest.mark.django_db

from django.db import IntegrityError
from django.db.models import ProtectedError
from django.core.exceptions import ValidationError as DjangoValidationError
import logging

from nsot import exc, models

from .fixtures import admin_user, user, site, transactional_db


def test_reparent_bug_issues_27(site):
    """
    Test for bug described at https://github.com/dropbox/nsot/issues/27
    """

    net_8  = models.Network.objects.create(site=site, cidr=u"10.0.0.0/8")
    net_31 = models.Network.objects.create(site=site, cidr=u"10.17.244.128/31")
    net_25 = models.Network.objects.create(site=site, cidr=u"10.16.1.0/25")

    net_8.refresh_from_db()
    net_31.refresh_from_db()
    net_25.refresh_from_db()

    assert net_8.parent_id is None
    assert net_31.parent_id == net_8.id
    assert net_25.parent_id == net_8.id



from __future__ import absolute_import, print_function

import sys

from cryptography.fernet import Fernet
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):
    help = (
        'Generate a URL-safe base64-encoded 32-byte key for use in '
        'settings.SECRET_KEY.'
    )

    def handle(self, **options):
        print(Fernet.generate_key())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from calendar import timegm
from cryptography.fernet import (Fernet, InvalidToken)
from custom_user.models import AbstractEmailUser
from django.db import models
from django.db.models.query_utils import Q
from django.conf import settings
from django.core.exceptions import ValidationError as DjangoValidationError
from django.core.validators import EmailValidator
from django_extensions.db.fields.json import JSONField
import ipaddress
import json
import logging
from operator import attrgetter
from polymorphic import PolymorphicModel
import re

from . import constants
from . import exc
from . import fields
from .util import generate_secret_key, parse_set_query


log = logging.getLogger(__name__)

RESOURCE_BY_IDX = (
    'Site', 'Network', 'Attribute', 'Permission', 'Device'
)
RESOURCE_BY_NAME = {
    obj_type: idx
    for idx, obj_type in enumerate(RESOURCE_BY_IDX)
}

CHANGE_EVENTS = ('Create', 'Update', 'Delete')
IP_VERSIONS = ('4', '6')


VALID_CHANGE_RESOURCES = set(RESOURCE_BY_IDX)
VALID_ATTRIBUTE_RESOURCES = set([
    'Network', 'Device',
])

ATTRIBUTE_RESOURCES = ('Network', 'Device')



class Site(models.Model):
    """A namespace for attribtues, devices, and networks."""
    name = models.CharField(max_length=255, unique=True)
    description = models.TextField(default='', blank=True)

    def __unicode__(self):
        return self.name

    def clean_name(self, value):
        if not value:
            raise exc.ValidationError({
                'name': 'This is a required field.'
            })
        return value

    def clean_fields(self, exclude=None):
        self.name = self.clean_name(self.name)

    def save(self, *args, **kwargs):
        self.full_clean()  # First validate fields are correct
        super(Site, self).save(*args, **kwargs)

    def to_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'description': self.description,
        }


class User(AbstractEmailUser):
    """A custom user object that utilizes email as the username."""
    secret_key = models.CharField(max_length=44, default=generate_secret_key)

    def get_permissions(self):
        permissions = []
        if self.is_staff or self.is_superuser:
            permissions.append('admin')
        sites = Site.objects.all()

        return {
            str(site.id): {
                'permissions': permissions,
                'site_id': site.id,
                'user_id': self.id
            }
            for site in sites
        }

    def rotate_secret_key(self):
        self.secret_key = generate_secret_key()
        self.save()

    def generate_auth_token(self):
        """Serialize user data and encrypt token."""
        # Serialize data
        data = json.dumps({'email': self.email})

        # Encrypt w/ servers's secret_key
        f = Fernet(bytes(settings.SECRET_KEY))
        auth_token = f.encrypt(bytes(data))
        return auth_token

    def verify_secret_key(self, secret_key):
        """Validate secret_key"""
        return secret_key == self.secret_key

    @classmethod
    def verify_auth_token(cls, email, auth_token, expiration=None):
        """Verify token and return a User object."""
        if expiration is None:
            expiration = settings.AUTH_TOKEN_EXPIRY

        # First we lookup the user by email
        query = cls.objects.filter(email=email)
        user = query.first()

        if user is None:
            log.debug('Invalid user when verifying token')
            raise exc.ValidationError({
                'auth_token': 'Invalid user when verifying token'
            })
            # return None  # Invalid user

        # Decrypt auth_token w/ user's secret_key
        f = Fernet(bytes(settings.SECRET_KEY))
        try:
            decrypted_data = f.decrypt(bytes(auth_token), ttl=expiration)
        except InvalidToken:
            log.debug('Invalid/expired auth_token when decrypting.')
            raise exc.ValidationError({
                'auth_token': 'Invalid/expired auth_token.'
            })
            # return None  # Invalid token

        # Deserialize data
        try:
            data = json.loads(decrypted_data)
        except ValueError:
            log.debug('Token could not be deserialized.')
            raise exc.ValidationError({
                'auth_token': 'Token could not be deserialized.'
            })
            # return None  # Valid token, but expired

        if email != data['email']:
            log.debug('Invalid user when deserializing.')
            raise exc.ValidationError({
                'auth_token': 'Invalid user when deserializing.'
            })
            # return None  # User email did not match payload
        return user

    def clean_email(self, value):
        validator = EmailValidator()
        try:
            validator(value)
        except DjangoValidationError as err:
            raise exc.ValidationError({
                'email': err.message
            })
        return value

    def clean_fields(self, exclude=None):
        self.email = self.clean_email(self.email)

    def save(self, *args, **kwargs):
        self.full_clean()
        super(User, self).save(*args, **kwargs)

    def to_dict(self, with_permissions=False, with_secret_key=False):
        out = [
            ('id', self.id),
            ('email', self.email),
        ]

        if with_secret_key:
            out.append(('secret_key', self.secret_key))

        if with_permissions:
            out.append(('permissions', self.get_permissions()))

        return dict(out)


class ResourceSetTheoryQuerySet(models.QuerySet):
    """
    Set theory QuerySet for Resource objects to add ``.set_query()`` method.

    For example::

        >>> qs = Network.objects.filter(network_address=u'10.0.0.0')
        >>> qs.set_query('owner=jathan +metro=lax')
    """
    def set_query(self, query, site_pk=None):
        objects = self
        if site_pk is not None:
            objects = objects.filter(site=site_pk)

        try:
            attributes = parse_set_query(query)
        except (ValueError, TypeError):
            attributes = []

        resource_name = self.model.__name__

        # Iterate a/v pairs and combine query results using MySQL-compatible
        # set operations w/ the ORM
        log.debug('QUERY [start]: objects = %r', objects)
        for action, name, value in attributes:
            attr = Attribute.objects.get(name=name, resource_name=resource_name)
            next_set = Q(
                attributes=Value.objects.filter(
                    attribute_id=attr.id, value=value
                ).values_list('id', flat=True)
            )

            # This is the MySQL-compatible manual implementation of set theory,
            # baby!
            if action == 'union':
                log.debug('SQL UNION')
                objects = (
                    objects | self.filter(next_set)
                )
            elif action == 'difference':
                log.debug('SQL DIFFERENCE')
                objects = objects.exclude(next_set)
            elif action == 'intersection':
                log.debug('SQL INTERSECTION')
                objects = objects.filter(next_set)
            else:
                raise exc.BadRequest('BAD SET QUERY: %r' % (action,))
            log.debug('QUERY [iter]: objects = %r', objects)

        # Gotta call .distinct() or we might get dupes.
        return objects.distinct()


class ResourceManager(models.Manager):
    """
    Manager for Resource objects that adds a ``.set_query()`` method.

    For example::

        >>> Network.objects.set_query('owner=jathan +metro=lax'}
        [<Device: foo-bar1>]
    """
    def get_queryset(self, *args, **kwargs):
        return ResourceSetTheoryQuerySet(self.model, using=self._db)

    def set_query(self, query, site_pk=None):
        return self.get_queryset().set_query(query, site_pk)


class Resource(PolymorphicModel):
    """Base for heirarchial Resource objects that may have attributes."""
    parent = models.ForeignKey(
        'self', blank=True, null=True, related_name='children', default=None,
        on_delete=models.PROTECT
    )
    _attributes = JSONField(null=False, blank=True)

    def __init__(self, *args, **kwargs):
        self._set_attributes = kwargs.pop('attributes', None)
        super(Resource, self).__init__(*args, **kwargs)

    # Implement .objects.set_query()
    objects = ResourceManager()

    @property
    def _resource_name(self):
        return self._meta.model_name.title()

    def _purge_attribute_index(self):
        self.attributes.all().delete()

    def get_attributes(self):
        """Return the JSON-encoded attributes as a dict."""
        return self._attributes

    def set_attributes(self, attributes, valid_attributes=None):
        """Validate and store the attributes dict as a JSON-encoded string."""
        log.debug('Resource.set_attributes() attributes = %r',
                  attributes)
        if not isinstance(attributes, dict):
            raise exc.ValidationError(
                'Expected dictionary but received {}'.format(type(attributes))
            )

        if valid_attributes is None:
            valid_attributes = Attribute.all_by_name(self._resource_name)
        log.debug('Resource.set_attributes() valid_attributes = %r',
                  valid_attributes)

        missing_attributes = {
            attribute.name for attribute in valid_attributes.itervalues()
            if attribute.required and attribute.name not in attributes
        }
        log.debug('Resource.set_attributes() missing_attributes = %r',
                  missing_attributes)

        if missing_attributes:
            names = ', '.join(missing_attributes)
            raise exc.ValidationError(
                'Missing required attributes: {}'.format(names)
            )

        inserts = []
        for name, value in attributes.iteritems():
            if name not in valid_attributes:
                raise exc.ValidationError(
                    'Attribute name ({}) does not exist.'.format(name)
                )

            if not isinstance(name, basestring):
                raise exc.ValidationError(
                    'Attribute names must be a string type.'
                )

            attribute = valid_attributes[name]
            inserts.extend(attribute.validate_value(value))

        # Purge all of our previously existing attribute values and recreate
        # them anew. This isn't exactly efficient.
        self._purge_attribute_index()
        for insert in inserts:
            Value.objects.create(
                resource=self, attribute_id=insert['attribute_id'],
                value=insert['value']
            )

        self.clean_attributes()

    def clean_attributes(self):
        """Make sure that attributes are saved as JSON."""
        attrs = {}
        for a in self.attributes.all():
            if a.attribute.multi:
                if a.name not in attrs:
                    attrs[a.name] = []
                attrs[a.name].append(a.value)
            else:
                attrs[a.name] = a.value
        self._attributes = attrs
        return attrs

    def clean_fields(self, exclude=None):
        self._attributes = self.clean_attributes()

    def save(self, *args, **kwargs):
        is_new = self.id is None  # Check if this is a new object.
        # self.full_clean()
        super(Resource, self).save(*args, **kwargs)

        # This is so that we can set the attributes on create/update, but if
        # the object is new, make sure that it doesn't persist if attributes
        # fail.
        if self._set_attributes is not None:
            try:
                # And set the attributes (if any)
                self.set_attributes(self._set_attributes)
            except exc.ValidationError:
                # If attributes fail validation, and I'm a new object, delete
                # myself and re-raise the error.
                if is_new:
                    self.delete()
                raise


class Device(Resource):
    """Represents a network device."""
    hostname = models.CharField(max_length=255, null=False, db_index=True)
    site = models.ForeignKey(
        Site, db_index=True, related_name='devices', on_delete=models.PROTECT
    )

    def __unicode__(self):
        return u'%s' % self.hostname

    class Meta:
        unique_together = ('site', 'hostname')
        index_together = unique_together

    def clean_hostname(self, value):
        if not value:
            raise exc.ValidationError({
                'hostname': 'Hostname must be non-zero length string.'
            })
        return value

    def clean_fields(self, exclude=None):
        self.hostname = self.clean_hostname(self.hostname)

    def save(self, *args, **kwargs):
        self.full_clean()
        super(Device, self).save(*args, **kwargs)

    def to_dict(self):
        return {
            'id': self.id,
            'site_id': self.site_id,
            'hostname': self.hostname,
            'attributes': self.get_attributes(),
        }


class Network(Resource):
    """Represents a subnet or ip address."""
    IP_VERSION_CHOICES = [(c, c) for c in IP_VERSIONS]

    network_address = fields.BinaryIPAddressField(
        max_length=16, null=False, db_index=True
    )
    broadcast_address = fields.BinaryIPAddressField(
        max_length=16, null=False, db_index=True
    )
    prefix_length = models.IntegerField(null=False, db_index=True)
    ip_version = models.CharField(
        max_length=1, null=False, db_index=True,
        choices=IP_VERSION_CHOICES
    )
    is_ip = models.BooleanField(null=False, default=False, db_index=True)
    site = models.ForeignKey(
        Site, db_index=True, related_name='networks', on_delete=models.PROTECT
    )

    def __init__(self, *args, **kwargs):
        self._cidr = kwargs.pop('cidr', None)
        super(Network, self).__init__(*args, **kwargs)

    def __unicode__(self):
        return self.cidr

    class Meta:
        unique_together = (
            'site', 'ip_version', 'network_address', 'prefix_length'
        )
        index_together = unique_together

    def supernets(self, direct=False, discover_mode=False, for_update=False):
        query = Network.objects.all()

        if self.parent is None and not discover_mode:
            return query.none()

        if discover_mode and direct:
            raise exc.ValidationError(
                'Direct is incompatible with discover_mode.'
            )

        if for_update:
            query = query.select_for_update()

        if direct:
            return query.filter(id=self.parent.id)

        return query.filter(
            site=self.site,
            is_ip=False,
            ip_version=self.ip_version,
            prefix_length__lt=self.prefix_length,
            network_address__lte=self.network_address,
            broadcast_address__gte=self.broadcast_address
        )

    def subnets(self, include_networks=True, include_ips=False, direct=False,
                for_update=False):
        query = Network.objects.all()

        if not any([include_networks, include_ips]) or self.is_ip:
            return query.none()

        if for_update:
            query = query.select_for_update()

        if not all([include_networks, include_ips]):
            if include_networks:
                query = query.filter(is_ip=False)
            if include_ips:
                query = query.filter(is_ip=True)

        if direct:
            return query.filter(parent__id=self.id)

        return query.filter(
            site=self.site,
            ip_version=self.ip_version,
            prefix_length__gt=self.prefix_length,
            network_address__gte=self.network_address,
            broadcast_address__lte=self.broadcast_address
        )

    @property
    def cidr(self):
        return u'%s/%s' % (self.network_address, self.prefix_length)

    @property
    def ip_network(self):
        return ipaddress.ip_network(self.cidr)

    def reparent_subnets(self):
        """
        Determine list of child nodes and set the parent to self.
        """
        query = Network.objects.select_for_update().filter(
            ~models.Q(id=self.id),  # Don't include yourself...
            parent_id=self.parent_id,
            prefix_length__gt=self.prefix_length,
            ip_version=self.ip_version,
            network_address__gte=self.network_address,
            broadcast_address__lte=self.broadcast_address
        )

        query.update(parent=self)

    def clean_fields(self, exclude=None):
        """This will enforce correct values on fields."""
        cidr = self._cidr
        if cidr is None:
            if self.network_address and self.prefix_length:
                cidr = u'%s/%s' % (self.network_address, self.prefix_length)

        if not cidr:
            msg = "Invalid CIDR: {}. Must be IPv4/IPv6 notation.".format(cidr)
            raise exc.ValidationError(msg)

        # Determine network properties
        network = cidr
        if isinstance(cidr, unicode):
            try:
                network = ipaddress.ip_network(cidr)
            except ValueError as err:
                raise exc.ValidationError({
                    'cidr': err.message
                })

        if network.network_address == network.broadcast_address:
            self.is_ip = True

        self.ip_version = str(network.version)
        self.network_address = unicode(network.network_address)
        self.broadcast_address = unicode(network.broadcast_address)
        self.prefix_length = network.prefixlen

    def save(self, *args, **kwargs):
        """This is stuff we want to happen upon save."""
        self.full_clean()  # First validate fields are correct

        for_update = kwargs.pop('for_update', False)

        # Calculate our supernets and determine if we require a parent.
        supernets = self.supernets(discover_mode=True, for_update=for_update)
        if supernets:
            parent = max(supernets, key=attrgetter('prefix_length'))
            self.parent = parent

        if self.parent is None and self.is_ip:
            raise exc.ValidationError('IP Address needs base network.')

        # Save, so we get an ID, and register our parent.
        super(Network, self).save(*args, **kwargs)

        # If we're not an IP, determine our subnets and reparent them.
        if not self.is_ip:
            self.reparent_subnets()

    def to_dict(self):
        return {
            'id': self.id,
            'parent_id': self.parent_id,
            'site_id': self.site_id,
            'is_ip': self.is_ip,
            'ip_version': self.ip_version,
            'network_address': self.network_address,
            'prefix_length': self.prefix_length,
            'attributes': self.get_attributes(),
        }


def validate_name(value):
    if not value:
        raise exc.ValidationError("Name is a required field.")

    if not constants.ATTRIBUTE_NAME.match(value):
        raise exc.ValidationError("Invalid name.")

    return value or False


class Attribute(models.Model):
    """Represents a flexible attribute for Resource objects."""
    # This is purposely not unique as there is a compound index with site_id.
    name = models.CharField(max_length=64, null=False, db_index=True)
    description = models.CharField(max_length=255, default='', null=False)

    # The resource must contain a key and value
    required = models.BooleanField(default=False, null=False)

    # In UIs this attribute will be displayed by default. Required implies
    # display.
    display = models.BooleanField(default=False, null=False)

    # Attribute values are expected as lists of strings.
    multi = models.BooleanField(default=False, null=False)

    constraints = JSONField(null=False, blank=True)

    site = models.ForeignKey(
        Site, db_index=True, related_name='attributes',
        on_delete=models.PROTECT
    )

    RESOURCE_CHOICES = [(c, c) for c in ATTRIBUTE_RESOURCES]

    resource_name = models.CharField(
        'Resource Name', max_length=20, null=False, db_index=True,
        choices=RESOURCE_CHOICES
    )

    def __unicode__(self):
        return u'%s %s' % (self.resource_name, self.name)

    class Meta:
        unique_together = ('site', 'resource_name', 'name')
        index_together = unique_together

    @classmethod
    def all_by_name(cls, resource_name=None):
        if resource_name is None:
            raise SyntaxError('You must provided a resource_name.')

        query = cls.objects.filter(resource_name=resource_name)

        return {
            attribute.name: attribute
            for attribute in query.all()
        }

    def clean_constraints(self, value):
        """Enforce formatting of constraints."""
        if not isinstance(value, dict):
            raise exc.ValidationError({
                'constraints': 'Expected dictionary but received {}.'.format(
                    type(value))
                })

        constraints = {
            'allow_empty': value.get('allow_empty', False),
            'pattern': value.get('pattern', ''),
            'valid_values': value.get('valid_values', []),
        }

        if not isinstance(constraints['allow_empty'], bool):
            raise exc.ValidationError({
                'constraints': 'allow_empty expected type bool.'
            })

        if not isinstance(constraints['pattern'], basestring):
            raise exc.ValidationError({
                'constraints': 'pattern expected type string.'
            })

        if not isinstance(constraints['valid_values'], list):
            raise exc.ValidationError({
                'constraints': 'valid_values expected type list.'
            })

        return constraints

    def clean_display(self, value):
        if self.required:
            return True
        return value

    def clean_resource_name(self, value):
        if value not in VALID_ATTRIBUTE_RESOURCES:
            raise exc.ValidationError({
                'resource_name': 'Invalid resource name: %r.' % value
            })
        return value

    def clean_name(self, value):
        if not value:
            raise exc.ValidationError({
                'name': 'This field is required.'
            })

        if not constants.ATTRIBUTE_NAME.match(value):
            raise exc.ValidationError({
                'name': 'Invalid name.'
            })

        return value or False

    def clean_fields(self, exclude=None):
        self.constraints = self.clean_constraints(self.constraints)
        self.display = self.clean_display(self.display)
        self.resource_name = self.clean_resource_name(self.resource_name)
        self.name = self.clean_name(self.name)

    def _validate_single_value(self, value, constraints=None):
        if not isinstance(value, basestring):
            raise exc.ValidationError({
                'value': 'Attribute values must be a string type'
            })

        if constraints is None:
            constraints = self.constraints

        allow_empty = constraints.get('allow_empty', False)
        if not allow_empty and not value:
            raise exc.ValidationError({
                'constraints': "Attribute {} doesn't allow empty values"
                .format(self.name)
            })

        pattern = constraints.get('pattern')
        if pattern and not re.match(pattern, value):
            raise exc.ValidationError({
                'pattern': "Attribute value {} for {} didn't match pattern: {}"
                .format(value, self.name, pattern)
            })

        valid_values = set(constraints.get('valid_values', []))
        if valid_values and value not in valid_values:
            raise exc.ValidationError(
                'Attribute value {} for {} not a valid value: {}'
                .format(value, self.name, ', '.join(valid_values))
            )

        return {
            'attribute_id': self.id,
            'value': value,
        }

    def validate_value(self, value):
        if self.multi:
            if not isinstance(value, list):
                raise exc.ValidationError({
                    'multi': 'Attribute values must be a list type'
                })
        else:
            value = [value]

        inserts = []
        # This does a deserialization so save the result
        constraints = self.constraints
        for val in value:
            inserts.append(self._validate_single_value(val, constraints))

        return inserts

    def save(self, *args, **kwargs):
        """Always enforce constraints."""
        self.full_clean()
        super(Attribute, self).save(*args, **kwargs)

    def to_dict(self):
        return {
            'id': self.id,
            'site_id': self.site_id,
            'description': self.description,
            'name': self.name,
            'resource_name': self.resource_name,
            'required': self.required,
            'display': self.display,
            'multi': self.multi,
            'constraints': self.constraints,
        }


class Value(models.Model):
    """Represents a value for an attribute attached to a Resource."""
    attribute = models.ForeignKey(
        Attribute, related_name='values', db_index=True,
        on_delete=models.PROTECT
    )
    value = models.CharField(
        max_length=255, null=False, blank=True,
        db_index=True
    )
    resource = models.ForeignKey(
        'Resource', related_name='attributes', db_index=True,
        blank=True
    )

    def __unicode__(self):
        return u'%r %s=%s' % (self.resource, self.name, self.value)

    class Meta:
        unique_together = ('attribute', 'value', 'resource')
        index_together = unique_together

    @property
    def name(self):
        return self.attribute.name

    @property
    def resource_name(self):
        return self.attribute.resource_name

    def clean_fields(self, exclude=None):
        log.debug('cleaning %s', self.name)
        # Make sure that the resource's type matches the resource name
        if self.resource_name != self.resource._resource_name:
            raise exc.ValidationError({
                'attribute': 'Invalid attribute type for this resource.'
            })

    def save(self, *args, **kwargs):
        self.full_clean()
        super(Value, self).save(*args, **kwargs)

    def to_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'value': self.value,
            'attribute': self.attribute.to_dict(),
            'resource_name': self.resource_name,
            'resource': self.resource.to_dict(),
        }


class Change(models.Model):
    """Record of all changes in NSoT."""
    EVENT_CHOICES = [(c, c) for c in CHANGE_EVENTS]
    RESOURCE_CHOICES = [(c, c) for c in VALID_CHANGE_RESOURCES]

    site = models.ForeignKey(Site, db_index=True, related_name='changes')
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name='changes', db_index=True
    )
    change_at = models.DateTimeField(auto_now_add=True, null=False)
    event = models.CharField(
        max_length=10, null=False, choices=EVENT_CHOICES,
    )
    resource_id = models.IntegerField('Resource ID', null=False)
    resource_name = models.CharField(
        'Resource Type', max_length=20, null=False, db_index=True,
        choices=RESOURCE_CHOICES
    )
    _resource = JSONField('Resource', null=False, blank=True)

    def __init__(self, *args, **kwargs):
        self._obj = kwargs.pop('obj', None)
        super(Change, self).__init__(*args, **kwargs)

    class Meta:
        get_latest_by = 'change_at'

    def __unicode__(self):
        return u'%s %s(%s)' % (self.event, self.resource_name,
                               self.resource_id)

    @property
    def resource(self):
        return self._resource

    def get_change_at(self):
        return timegm(self.change_at.timetuple())
    get_change_at.short_description = 'Change At'

    @classmethod
    def get_serializer_for_resource(cls, resource_name):
        from .api import serializers
        serializer_class = resource_name + 'Serializer'
        return getattr(serializers, serializer_class)

    def clean_event(self, value):
        if value not in CHANGE_EVENTS:
            raise exc.ValidationError('Invalid change event: %r.' % value)
        return value

    def clean_resource_name(self, value):
        if value not in VALID_CHANGE_RESOURCES:
            raise exc.ValidationError('Invalid resource name: %r.' % value)
        return value

    def clean_fields(self, exclude=None):
        """This will populate the change fields from the incoming object."""
        obj = self._obj
        if obj is None:
            return None

        self.event = self.clean_event(self.event)
        self.resource_name = self.clean_resource_name(obj.__class__.__name__)
        self.resource_id = obj.id

        # Site doesn't have an id to itself, so if obj is a Site, use it.
        self.site = obj if isinstance(obj, Site) else obj.site

        serializer_class = self.get_serializer_for_resource(self.resource_name)
        serializer = serializer_class(obj)
        self._resource = serializer.data

    def save(self, *args, **kwargs):
        self.full_clean()  # First validate fields are correct
        super(Change, self).save(*args, **kwargs)

    def to_dict(self):
        resource = None
        if self.resource is not None:
            resource = self.resource

        return {
            'id': self.id,
            'site': self.site.to_dict(),
            'user': self.user.to_dict(),
            'change_at': timegm(self.change_at.timetuple()),
            'event': self.event,
            'resource_name': self.resource_name,
            'resource_id': self.resource_id,
            'resource': resource,
        }



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest

# Allow everything in there to access the DB
pytestmark = pytest.mark.django_db

import copy
from django.core.urlresolvers import reverse
import json
import logging
from rest_framework import status
import requests


from .fixtures import live_server, client, user, site
from .util import (
    assert_created, assert_error, assert_success, assert_deleted, load_json,
    Client, load
)


log = logging.getLogger(__name__)


def test_no_user(live_server, site):
    site_uri = site.list_uri()
    url = '{}{}'.format(live_server.url, site_uri)

    assert_error(requests.get(url), status.HTTP_401_UNAUTHORIZED)


def test_valid_user(live_server):
    url = '{}/api/sites/'.format(live_server.url)
    headers = {'X-NSoT-Email': 'gary@localhost'}

    assert_success(
        requests.get(url, headers=headers),
        {'sites': [], 'limit': None, 'offset': 0, 'total': 0}
    )


def test_invalid_user(live_server):
    url = '{}/api/sites/'.format(live_server.url)
    headers = {'X-NSoT-Email': 'gary'}

    assert_error(
        requests.get(url, headers=headers),
        status.HTTP_400_BAD_REQUEST
    )


def test_get_auth_token_valid(live_server, user):
    """Test that an auth_token can be generated."""
    auth_uri = reverse('authenticate')
    url = '{}{}'.format(live_server.url, auth_uri)

    headers={'Content-Type': 'application/json'}
    payload = {'email': user.email, 'secret_key': user.secret_key}
    data = json.dumps(payload)

    resp = requests.post(url, headers=headers, data=data)

    assert_success(resp, resp.json()['data'])


def test_get_auth_token_invalid(live_server, user):
    """Test that an auth_token fails w/ a bad secret key."""
    auth_uri = reverse('authenticate')
    url = '{}{}'.format(live_server.url, auth_uri)

    headers={'Content-Type': 'application/json'}
    payload = {'email': user.email, 'secret_key': 'bogus'}
    data = json.dumps(payload)

    assert_error(
        requests.post(url, headers=headers, data=data),
        status.HTTP_401_UNAUTHORIZED
    )


def test_get_auth_token_missing(live_server, user):
    """Test that missing payload results in a 401."""
    auth_uri = reverse('authenticate')
    url = '{}{}'.format(live_server.url, auth_uri)
    headers={'Content-Type': 'application/json'}

    assert_error(
        requests.post(url, headers=headers, data=''),
        status.HTTP_401_UNAUTHORIZED
    )


def test_verify_auth_token_invalid(live_server, user):
    """Test that an auth_token is NOT valid."""
    verify_url = '{}{}'.format(live_server.url, reverse('verify_token'))

    headers = {
        'Content-Type': 'application/json',
        'Authorization': 'AuthToken {}:{}'.format(user.email, 'bogus')
    }

    assert_error(
        requests.post(verify_url, headers=headers),
        status.HTTP_400_BAD_REQUEST
    )


def test_verify_auth_token_valid(live_server, site, user):
    """Test that an auth_token is valid."""
    auth_url = '{}{}'.format(live_server.url, reverse('authenticate'))
    verify_url = '{}{}'.format(live_server.url, reverse('verify_token'))

    headers = {'Content-Type': 'application/json'}
    payload = {'email': user.email, 'secret_key': user.secret_key}
    data = json.dumps(payload)

    auth_resp = requests.post(auth_url, headers=headers, data=data)
    auth_token = auth_resp.json()['data']['auth_token']

    headers.update({
        'Authorization': 'AuthToken {}:{}'.format(user.email, auth_token)
    })

    assert_success(requests.post(verify_url, headers=headers), True)


def test_valid_auth_token(live_server, user):
    """Test that a GET can be performed to a resource using auth_token."""
    auth_url = '{}{}'.format(live_server.url, reverse('authenticate'))
    site_url = '{}{}'.format(live_server.url, reverse('site-list'))

    headers = {'Content-Type': 'application/json'}
    payload = {'email': user.email, 'secret_key': user.secret_key}
    data = json.dumps(payload)

    auth_resp = requests.post(auth_url, headers=headers, data=data)
    auth_token = auth_resp.json()['data']['auth_token']

    headers.update({
        'Authorization': 'AuthToken {}:{}'.format(user.email, auth_token)
    })

    # Allow the user to login to the API
    user.is_staff = True
    user.save()

    assert_success(
        requests.get(site_url, headers=headers),
        {'sites': [], 'limit': None, 'offset': 0, 'total': 0}
    )



from django.core.urlresolvers import reverse
from hashlib import sha1
import json
import os
from rest_framework import status
import requests
from urlparse import urlparse


def _deep_sort(obj):
    """Sort the items in an object so comparisons succeed."""
    if isinstance(obj, dict):
        return {
            key: _deep_sort(value)
            for key, value in obj.iteritems()
        }
    elif isinstance(obj, list):
        return sorted(_deep_sort(elem) for elem in obj)
    return obj


def assert_error(response, code):
    """Assert a response resulted in an error."""
    output = response.json()
    assert output['status'] == 'error'
    assert output['error']['code'] == code


def assert_success(response, data=None, ignore_order=True):
    """Assert a response resulted in an success."""
    output = response.json()
    assert response.status_code == status.HTTP_200_OK
    assert output['status'] == 'ok'

    if data is not None:
        print 'OUTPUT_DATA = %r' % (output['data'],)
        print 'QUERY_DATA = %r' % (data,)
        if ignore_order:
            assert _deep_sort(output['data']) == _deep_sort(data)
        else:
            assert output['data'] == data


def assert_created(response, location, data=None):
    """Assert 201 CREATED."""
    output = response.json()
    assert response.status_code == status.HTTP_201_CREATED

    resp_location = response.headers.get('Location')
    if resp_location is not None:
        url = urlparse(resp_location)
        resp_location = url.path

    assert resp_location == location
    assert output['status'] == 'ok'

    if data is not None:
        assert output['data'] == data


def assert_deleted(response):
    assert response.status_code == status.HTTP_204_NO_CONTENT

    # TODO(jathan): If we want the deleted object to be returned, then this
    # test will have to be updated.
    # output = response.json()
    # assert output["status"] == "ok"


class Client(object):
    def __init__(self, live_server, user="admin"):
        self.live_server = live_server
        self.user = "{}@localhost".format(user)
        self.session = requests.Session()

    @property
    def base_url(self):
        return self.live_server.url

    # def request(self, method, url, user="admin", **kwargs):
    def request(self, method, url, user="admin", **kwargs):

        headers = {
            "X-NSoT-Email": self.user
        }

        if method.lower() in ("put", "post"):
            headers["Content-type"] = "application/json"

        from betamax import Betamax
        cassette = sha1(url).hexdigest()  # SHA1 the URI!
        with Betamax(self.session).use_cassette(cassette):
            return self.session.request(
                method, self.base_url + url,
                headers=headers, **kwargs
            )

        '''
        return self.session.request(
            method, self.base_url + url,
            headers=headers, **kwargs
        )
        '''

    def get(self, url, **kwargs):
        return self.request("GET", url, **kwargs)

    def post(self, url, **kwargs):
        return self.request("POST", url, **kwargs)

    def put(self, url, **kwargs):
        return self.request("PUT", url, **kwargs)

    def delete(self, url, **kwargs):
        return self.request("DELETE", url, **kwargs)

    def create(self, url, **kwargs):
        return self.post(url, data=json.dumps(kwargs))

    def update(self, url, **kwargs):
        return self.put(url, data=json.dumps(kwargs))

    def retrieve(self, url, **kwargs):
        return self.get(url, params=kwargs)


def load_json(relpath):
    """
    Load JSON files relative to this directory.

    Files are loaded from the 'data' directory. So for example for
    ``/path/to/data/devices/foo.json`` the ``relpath`` would be
    ``devices/foo.json``.

    :param relpath:
        Relative path to our directory's "data" dir
    """
    our_path = os.path.dirname(os.path.abspath(__file__))
    data_dir = os.path.join(our_path, 'data')
    filepath = os.path.join(data_dir, relpath)
    with open(filepath, 'rb') as f:
        return json.load(f)


def load(relpath):
    our_path = os.path.dirname(os.path.abspath(__file__))
    data_dir = os.path.join(our_path, 'data')
    filepath = os.path.join(data_dir, relpath)
    with open(filepath, 'rb') as f:
        return f.read()


def filter_devices(devices, wanted):
    """
    Return a list of desired Device objects.

    :param devices:
        list of Device dicts

    :param wanted:
        list of cidrs you want
    """
    return [d for d in devices if d['hostname'] in wanted]


def filter_networks(networks, wanted):
    """
    Return a list of desired Network objects.

    :param networks:
        list of Network dicts

    :param wanted:
        list of cidrs you want
    """
    return [
        n for n in networks if '%s/%s' % (n['network_address'],
        n['prefix_length']) in wanted
    ]


class TestSite(object):
    def __init__(self, data):
        self._data = data
        self.__dict__.update(**data)

    def list_uri(self, name=None, site_id=None):
        """Return a list URL like /api/sites/ or /api/sites/1/devices/"""
        args = []
        if name is None:
            name = 'site'

        if site_id is None:
            site_id = self.id

        if site_id and name != 'site':
            args.append(site_id)

        return reverse(name + '-list', args=args)

    def detail_uri(self, name=None, site_id=None, id=None):
        """Return a detail URL like /api/sites/1/ or /api/sites/1/devices/1/"""
        if name is None:
            name = 'site'

        if site_id is None:
            site_id = self.id
        args = [site_id]

        if id is not None:
            args.append(id)

        return reverse(name + '-detail', args=args)

    def query_uri(self, name, site_id=None):
        """Return a query URL like /api/sites/1/devices/query/"""
        if site_id is None:
            site_id = self.id
        args = [site_id]

        return reverse(name + '-query', args=args)

    def __repr__(self):
        return '<TestSite: %s (%s)>' % (self.name, self.id)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.core.handlers.wsgi import STATUS_CODE_TEXT as status_text
from django.shortcuts import render
from django.views.generic import TemplateView
import logging


log = logging.getLogger(__name__)


class FeView(TemplateView):
    """
    Front-end UI view that hands-off rendering to Angular.js.

    Any additional context needed to be passed to the templates, should be added
    in ``.get_context_data()``.
    """
    template_name = 'ui/app.html'

    def get_context_data(self, **kwargs):
        from nsot import __version__
        context = super(FeView, self).get_context_data(**kwargs)
        context['NSOT_VERSION'] = __version__
        return context


def render_error(request, status_code, template_name='ui/error.html'):
    """Generic base for rendering error pages."""
    message = status_text.get(status_code)
    context = {'code': status_code, 'message': message}
    return render(request, template_name, context, status=status_code)


def handle400(request):
    """Handler for 400."""
    return render_error(request, 400)


def handle403(request):
    """Handler for 403."""
    return render_error(request, 403)


def handle404(request):
    """Handler for 404."""
    return render_error(request, 404)


def handle500(request):
    """Handler for 500."""
    return render_error(request, 500)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import nsot.util
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
import nsot.fields
import django_extensions.db.fields.json


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('auth', '0006_require_contenttypes_0002'),
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(null=True, verbose_name='last login', blank=True)),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('email', models.EmailField(unique=True, max_length=255, verbose_name='email address', db_index=True)),
                ('is_staff', models.BooleanField(default=False, help_text='Designates whether the user can log into this admin site.', verbose_name='staff status')),
                ('is_active', models.BooleanField(default=True, help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.', verbose_name='active')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now, verbose_name='date joined')),
                ('secret_key', models.CharField(default=nsot.util.generate_secret_key, max_length=44)),
                ('groups', models.ManyToManyField(related_query_name='user', related_name='user_set', to='auth.Group', blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', verbose_name='groups')),
                ('user_permissions', models.ManyToManyField(related_query_name='user', related_name='user_set', to='auth.Permission', blank=True, help_text='Specific permissions for this user.', verbose_name='user permissions')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'user',
                'verbose_name_plural': 'users',
            },
        ),
        migrations.CreateModel(
            name='Attribute',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=64, db_index=True)),
                ('description', models.CharField(default='', max_length=255)),
                ('required', models.BooleanField(default=False)),
                ('display', models.BooleanField(default=False)),
                ('multi', models.BooleanField(default=False)),
                ('constraints', django_extensions.db.fields.json.JSONField(blank=True)),
                ('resource_name', models.CharField(db_index=True, max_length=20, verbose_name='Resource Name', choices=[('Network', 'Network'), ('Device', 'Device')])),
            ],
        ),
        migrations.CreateModel(
            name='Change',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('change_at', models.DateTimeField(auto_now_add=True)),
                ('event', models.CharField(max_length=10, choices=[('Create', 'Create'), ('Update', 'Update'), ('Delete', 'Delete')])),
                ('resource_id', models.IntegerField(verbose_name='Resource ID')),
                ('resource_name', models.CharField(db_index=True, max_length=20, verbose_name='Resource Type', choices=[('Device', 'Device'), ('Attribute', 'Attribute'), ('Permission', 'Permission'), ('Site', 'Site'), ('Network', 'Network')])),
                ('_resource', django_extensions.db.fields.json.JSONField(verbose_name='Resource', blank=True)),
            ],
            options={
                'get_latest_by': 'change_at',
            },
        ),
        migrations.CreateModel(
            name='Resource',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('_attributes', django_extensions.db.fields.json.JSONField(blank=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Site',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(unique=True, max_length=255)),
                ('description', models.TextField(default='', blank=True)),
            ],
        ),
        migrations.CreateModel(
            name='Value',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('value', models.CharField(db_index=True, max_length=255, blank=True)),
                ('attribute', models.ForeignKey(related_name='values', on_delete=django.db.models.deletion.PROTECT, to='nsot.Attribute')),
            ],
        ),
        migrations.CreateModel(
            name='Device',
            fields=[
                ('resource_ptr', models.OneToOneField(parent_link=True, auto_created=True, primary_key=True, serialize=False, to='nsot.Resource')),
                ('hostname', models.CharField(max_length=255, db_index=True)),
                ('site', models.ForeignKey(related_name='devices', on_delete=django.db.models.deletion.PROTECT, to='nsot.Site')),
            ],
            bases=('nsot.resource',),
        ),
        migrations.CreateModel(
            name='Network',
            fields=[
                ('resource_ptr', models.OneToOneField(parent_link=True, auto_created=True, primary_key=True, serialize=False, to='nsot.Resource')),
                ('network_address', nsot.fields.BinaryIPAddressField(max_length=16, db_index=True)),
                ('broadcast_address', nsot.fields.BinaryIPAddressField(max_length=16, db_index=True)),
                ('prefix_length', models.IntegerField(db_index=True)),
                ('ip_version', models.CharField(db_index=True, max_length=1, choices=[('4', '4'), ('6', '6')])),
                ('is_ip', models.BooleanField(default=False, db_index=True)),
                ('site', models.ForeignKey(related_name='networks', on_delete=django.db.models.deletion.PROTECT, to='nsot.Site')),
            ],
            bases=('nsot.resource',),
        ),
        migrations.AddField(
            model_name='value',
            name='resource',
            field=models.ForeignKey(related_name='attributes', blank=True, to='nsot.Resource'),
        ),
        migrations.AddField(
            model_name='resource',
            name='parent',
            field=models.ForeignKey(related_name='children', on_delete=django.db.models.deletion.PROTECT, default=None, blank=True, to='nsot.Resource', null=True),
        ),
        migrations.AddField(
            model_name='resource',
            name='polymorphic_ctype',
            field=models.ForeignKey(related_name='polymorphic_nsot.resource_set+', editable=False, to='contenttypes.ContentType', null=True),
        ),
        migrations.AddField(
            model_name='change',
            name='site',
            field=models.ForeignKey(related_name='changes', to='nsot.Site'),
        ),
        migrations.AddField(
            model_name='change',
            name='user',
            field=models.ForeignKey(related_name='changes', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='attribute',
            name='site',
            field=models.ForeignKey(related_name='attributes', on_delete=django.db.models.deletion.PROTECT, to='nsot.Site'),
        ),
        migrations.AlterUniqueTogether(
            name='value',
            unique_together=set([('attribute', 'value', 'resource')]),
        ),
        migrations.AlterIndexTogether(
            name='value',
            index_together=set([('attribute', 'value', 'resource')]),
        ),
        migrations.AlterUniqueTogether(
            name='attribute',
            unique_together=set([('site', 'resource_name', 'name')]),
        ),
        migrations.AlterIndexTogether(
            name='attribute',
            index_together=set([('site', 'resource_name', 'name')]),
        ),
        migrations.AlterUniqueTogether(
            name='network',
            unique_together=set([('site', 'ip_version', 'network_address', 'prefix_length')]),
        ),
        migrations.AlterIndexTogether(
            name='network',
            index_together=set([('site', 'ip_version', 'network_address', 'prefix_length')]),
        ),
        migrations.AlterUniqueTogether(
            name='device',
            unique_together=set([('site', 'hostname')]),
        ),
        migrations.AlterIndexTogether(
            name='device',
            index_together=set([('site', 'hostname')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest

# Allow everything in there to access the DB
pytestmark = pytest.mark.django_db

import copy
from django.core.urlresolvers import reverse
import json
import logging
from rest_framework import status


from .fixtures import live_server, client, user, site
from .util import (
    assert_created, assert_error, assert_success, assert_deleted, load_json,
    Client, load, filter_networks
)


log = logging.getLogger(__name__)


def test_network_bug_issues_34(client, site):
    """Test set queries for Networks."""

    # URIs
    attr_uri = site.list_uri('attribute')
    net_uri = site.list_uri('network')

    # Pre-load the attributes
    client.post(attr_uri, data=load('attributes.json'))

    # Populate the network objects and retreive them for testing.
    client.post(net_uri, data=load('networks.json'))
    net_resp = client.retrieve(net_uri, include_ips=True)
    net_out = net_resp.json()['data']
    networks = net_out['networks']

    # Filter networks w/ attribute hostname=foo-bar1
    expected = copy.deepcopy(net_out)
    wanted = ['192.168.0.0/24', '192.168.0.0/25']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'limit': None, 'offset': 0, 'total': len(wanted)})

    assert_success(
        client.retrieve(net_uri, attributes='hostname=foo-bar1'),
        expected
    )

    # Filter networks w/ attribute hostname=foo-bar1, including IPs
    wanted = ['192.168.0.1/32', '192.168.0.0/24', '192.168.0.0/25']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'total': len(wanted)})

    assert_success(
        client.retrieve(
            net_uri, attributes='hostname=foo-bar1', include_ips=True
        ),
        expected
    )



"""
Project-wide utilities.
"""

import collections
from cryptography.fernet import Fernet
import logging
from logan.runner import run_app, configure_app


log = logging.getLogger(__name__)

_TRUTHY = set([
    'true', 'yes', 'on', '1', ''
])


def qpbool(arg):
    return str(arg).lower() in _TRUTHY


def normalize_auth_header(header):
    return 'HTTP_' + header.upper().replace('-', '_')


def generate_secret_key():
    """Return a secret key suitable for use w/ Fernet."""
    return Fernet.generate_key()


#: Namedtuple for resultant items from ``parse_set_query()``
SetQuery = collections.namedtuple('SetQuery', 'action name value')


def parse_set_query(query):
    """
    Parse a representation of set operations for attribute/value pairs into
    (action, name, value) and return a list of ``SetQuery`` objects.

    Computes left-to-right evaluation, where the first character indicates the
    set operation:

    + "+" indicates a union
    + "-" indicates a difference
    + no marker indicates an intersection

    For example::

        >>> parse_set_query('+owner=team-networking')
        [SetQuery(action='union', name='owner', value='team-networking')]
        >>> parse_set_query('foo=bar')
        [SetQuery(action='intersection', name='foo', value='bar')]
        >>> parse_set_query('foo=bar -owner=team-networking')
        [SetQuery(action='intersection', name='foo', value='bar'),
         SetQuery(action='difference', name='owner', value='team-networking')]

    :param query:
        Set query string
    """
    log.debug('Incoming query = %r' % (query,))
    queries = query.split()

    attributes = []
    for q in queries:
        if q.startswith('+'):
            action = 'union'
            q = q[1:]
        elif q.startswith('-'):
            action = 'difference'
            q = q[1:]
        else:
            action = 'intersection'

        name, _, value = q.partition('=')
        attributes.append(SetQuery(action, name, value))

    log.debug('Outgoing attributes = %r' % (attributes,))
    return attributes


#: Configuration template emitted when a user runs ``nsot-server init``.
CONFIG_TEMPLATE = '''
"""
This configuration file is just Python code. You may override any global
defaults by specifying them here.

For more information on this file, see
https://docs.djangoproject.com/en/1.8/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.8/ref/settings/
"""
from nsot.conf.settings import *

import os.path

# Path where the config is found.
CONF_ROOT = os.path.dirname(__file__)

# A boolean that turns on/off debug mode. Never deploy a site into production
# with DEBUG turned on.
# Default: False
DEBUG = False

############
# Database #
############
# https://docs.djangoproject.com/en/dev/ref/settings/#databases
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(CONF_ROOT, 'nsot.sqlite3'),
        'USER': 'nsot',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
    }
}

###############
# Application #
###############

# The address on which the application will listen.
# Default: localhost
NSOT_HOST = 'localhost'

# The port on which the application will be accessed.
# Default: 8990
NSOT_PORT = 8990

# If True, serve static files directly from the app.
# Default: True
SERVE_STATIC_FILES = True

############
# Security #
############

# A URL-safe base64-encoded 32-byte key. This must be kept secret. Anyone with
# this key is able to create and read messages. This key is used for
# encryption/decryption of sessions and auth tokens.
SECRET_KEY = %(secret_key)r

# Header to check for Authenticated Email. This is intended for use behind an
# authenticating reverse proxy.
USER_AUTH_HEADER = 'X-NSoT-Email'

# The age, in seconds, until an AuthToken granted by the API will expire.
# Default: 600
AUTH_TOKEN_EXPIRY = 600  # 10 minutes

# A list of strings representing the host/domain names that this Django site can
# serve. This is a security measure to prevent an attacker from poisoning caches
# and triggering password reset emails with links to malicious hosts by
# submitting requests with a fake HTTP Host header, which is possible even under
# many seemingly-safe web server configurations.
# https://docs.djangoproject.com/en/1.8/ref/settings/#allowed-hosts
ALLOWED_HOSTS = ['*']
'''


def generate_settings(config_template=None):
    """Used to emit a config template."""
    if config_template is None:
        config_template = CONFIG_TEMPLATE.strip()

    secret_key = generate_secret_key()
    return config_template % dict(secret_key=secret_key)


def initialize_app(config):
    """Actions to be performed prior to creating the Application object."""
    USE_GEVENT = True  # Hard-coding gevent for now.

    if USE_GEVENT:
        # TODO(jathan): We need to keep an eye on this. If we run into any race
        # conditions w/ database updates, this should be the first place we
        # look.
        from django.db import connections
        connections['default'].allow_thread_sharing = True


def main():
    """CLI application used to manage NSoT."""
    run_app(
        project='nsot',
        default_config_path='~/.nsot/nsot.conf.py',
        default_settings='nsot.conf.settings',
        settings_initializer=generate_settings,
        settings_envvar='NSOT_CONF',
        initializer=initialize_app,
    )


if __name__ == '__main__':
    main()



from __future__ import unicode_literals

from collections import OrderedDict
from django.template import Context, loader
import logging
from rest_framework import pagination
from rest_framework.response import Response


log = logging.getLogger(__name__)


class CustomPagination(pagination.LimitOffsetPagination):
    """Custom pagination that always shows pagination controls in list view."""
    def paginate_queryset(self, queryset, request, view=None):
        self.limit = self.get_limit(request)
        # This is so we can always display pagination without having to specify
        # an upper limite.
        # if self.limit is None:
        #     return None

        self.offset = self.get_offset(request)
        self.count = pagination._get_count(queryset)
        self.request = request
        if self.count > self.limit and self.template is not None:
            self.display_page_controls = True

        # If we have a limit, slice it
        if self.limit:
            return list(queryset[self.offset:self.offset + self.limit])
        # Otherwise only slice from the offset
        else:
            return list(queryset[self.offset:])

    def get_next_link(self):
        if not self.limit:
            return None
        return super(CustomPagination, self).get_next_link()

    def get_previous_link(self):
        if not self.limit:
            return None
        return super(CustomPagination, self).get_previous_link()

    def get_html_context(self):
        if self.limit is None:
            return {}
        return super(CustomPagination, self).get_html_context()

    def get_paginated_response(self, data, result_key=None):
        """
        Custom pagination response that excludes next/previous.

        :param data:
            Paginated data

        :param result_key:
            If set, use this as the key to label the results data.
        """
        log.debug('request URI = %s' % self.request.build_absolute_uri())
        log.debug('request path = %s' % self.request.path)
        log.debug('request path_info = %s' % self.request.path_info)

        # If path is '/api/sites/1/attributes/', this is 'attributes'
        if result_key is None:
            result_key = self.request.path.rstrip('/').split('/')[-1]
        return Response(
            OrderedDict([
                ('status', 'ok'),
                ('data', OrderedDict(
                [
                    ('total', self.count),
                    ('limit', self.limit),
                    ('offset', self.offset),
                    # ('next', self.get_next_link()),
                    # ('previous', self.get_previous_link()),
                    (result_key, data),
                    #('results', data),  # Generic 'results' key
                ]))
            ])
        )



from datetime import datetime
import logging
from time import time


class LoggingMiddleware(object):
    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def process_request(self, request):
        request.timer = time()
        return None

    def process_response(self, request, response):
        self.logger.info(
            '%s INFO %s %s %s %.2fms',
            datetime.now(),
            response.status_code,
            request.method,
            request.get_full_path(),
            (time() - request.timer) * 1000  # ms
        )
        return response



"""
This configuration file is just Python code. You may override any global
defaults by specifying them here.

For more information on this file, see
https://docs.djangoproject.com/en/1.8/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.8/ref/settings/
"""

from nsot.conf.settings import *
import os.path


# Path where the config is found.
CONF_ROOT = os.path.dirname(__file__)

# A boolean that turns on/off debug mode. Never deploy a site into production
# with DEBUG turned on.
# Default: False
DEBUG = False

############
# Database #
############
# https://docs.djangoproject.com/en/dev/ref/settings/#databases
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(CONF_ROOT, 'nsot.sqlite3'),
        'USER': 'nsot',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
    }
}

###############
# Application #
###############

# The address on which the application will listen.
# Default: localhost
NSOT_HOST = 'localhost'

# The port on which the application will be accessed.
# Default: 8990
NSOT_PORT = 8990

# If True, serve static files directly from the app.
# Default: True
SERVE_STATIC_FILES = True

############
# Security #
############

# A URL-safe base64-encoded 32-byte key. This must be kept secret. Anyone with
# this key is able to create and read messages. This key is used for
# encryption/decryption of sessions and auth tokens. A unique key is randomly
# generated for you when you utilize ``nsot-server init``
# https://cryptography.io/en/latest/fernet/#cryptography.fernet.Fernet.generate_key
SECRET_KEY = u'fMK68NKgazLCjjTXjDtthhoRUS8IV4lwD-9G7iVd2Xs='

# Header to check for Authenticated Email. This is intended for use behind an
# authenticating reverse proxy.
USER_AUTH_HEADER = 'X-NSoT-Email'

# The age, in seconds, until an AuthToken granted by the API will expire.
# Default: 600
AUTH_TOKEN_EXPIRY = 600  # 10 minutes

# A list of strings representing the host/domain names that this Django site can
# serve. This is a security measure to prevent an attacker from poisoning caches
# and triggering password reset emails with links to malicious hosts by
# submitting requests with a fake HTTP Host header, which is possible even under
# many seemingly-safe web server configurations.
# https://docs.djangoproject.com/en/1.8/ref/settings/#allowed-hosts
ALLOWED_HOSTS = ['*']



from __future__ import unicode_literals

from custom_user.admin import EmailUserAdmin
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib import admin
from django.utils.translation import ugettext_lazy as _
from polymorphic.admin import (PolymorphicParentModelAdmin,
                               PolymorphicChildModelAdmin)

from . import models


# We don't want to use the Group model for now.
admin.site.unregister(Group)


# Register our custom User model
class UserAdmin(EmailUserAdmin):
    fieldsets = (
        (None, {
            'fields': ('email', 'secret_key'),
        }),
        (_('Permissions'), {
            'fields': ('is_active', 'is_staff', 'is_superuser'),
                       #  'groups', 'user_permissions'),
        }),
        (_('Important dates'), {
            'fields': ('last_login', 'date_joined')
        }),
    )
admin.site.register(get_user_model(), UserAdmin)



class SiteAdmin(admin.ModelAdmin):
    list_display = ('name', 'description')
    list_filter = ('name',)
admin.site.register(models.Site, SiteAdmin)


class AttributeAdmin(admin.ModelAdmin):
    list_display = ('name', 'resource_name', 'description', 'required', 'display',
                    'multi', 'site')
    list_filter = ('name', 'resource_name', 'required', 'multi', 'site')
admin.site.register(models.Attribute, AttributeAdmin)


class ValueAdmin(admin.ModelAdmin):
    list_display = ('name', 'value', 'resource')
    list_filter = ('attribute', 'value')
admin.site.register(models.Value, ValueAdmin)


class ChangeAdmin(admin.ModelAdmin):
    list_display = ('id', 'user', 'event', 'resource_name', 'resource_id',
                    'get_change_at', 'resource_name', 'site')
    list_filter = ('event', 'site')
admin.site.register(models.Change, ChangeAdmin)


class BaseChildAdmin(PolymorphicChildModelAdmin):
    base_model = models.Resource

    base_fieldsets = (None, {
        'fields': ('id', 'site'),
    })


class DeviceAdmin(PolymorphicChildModelAdmin):
    base_model = models.Device
    list_display = ('hostname', 'site')
    list_filter = ('hostname', 'site')

    fields = list_display
admin.site.register(models.Device, DeviceAdmin)


class NetworkAdmin(BaseChildAdmin):
    base_model = models.Network
    mptt_level_indent = 10
    mptt_indent_field = 'cidr'
    list_display = ('cidr', 'network_address', 'prefix_length', 'ip_version',
                    'is_ip', 'parent', 'site')
    list_filter = ('prefix_length', 'is_ip', 'ip_version', 'site')

    def get_cidr(self):
        return '%s/%s' % (self.network_address, self.prefix_length)

    fields = ('network_address', 'broadcast_address', 'prefix_length',
              'ip_version', 'is_ip', 'parent', 'site')
admin.site.register(models.Network, NetworkAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest

# Allow everything in here to access the DB
pytestmark = pytest.mark.django_db

from django.core.urlresolvers import reverse
import logging
from rest_framework import status

from .fixtures import client
from .util import (
    assert_created, assert_deleted, assert_error, assert_success, TestSite
)


log = logging.getLogger(__name__)


def test_malformed(client):
    url = reverse('site-list')
    response = client.post(url, data='Non-JSON')

    assert_error(response, status.HTTP_400_BAD_REQUEST)


def test_creation(client):
    site_uri = reverse('site-list')

    assert_success(client.get(site_uri), {
        'sites': [],
        'limit': None,
        'offset': 0,
        'total': 0
    })

    # Create a Site
    site1_resp = client.create(site_uri, name='Test Site')
    site1 = TestSite(site1_resp.json()['data']['site'])
    site1_obj_uri = site1.detail_uri()
    assert_created(site1_resp, site1_obj_uri)

    # Try (and fail) to create another Site with the same name
    # TODO(jathan): Unique failures might need to be 409 vs. 400?
    assert_error(
        client.create(site_uri, name='Test Site'),
        # status.HTTP_409_CONFLICT
        status.HTTP_400_BAD_REQUEST
    )

    # Create another Site with a slightly different name
    site2_resp = client.create(site_uri, name='Test Site 2')
    site2 = TestSite(site2_resp.json()['data']['site'])
    site2_obj_uri = site2.detail_uri()
    assert_created(site2_resp, site2_obj_uri)

    # And now retrieve the first Site by name.
    expected = site1_resp.json()['data']
    expected['sites'] = [expected.pop('site')]
    expected.update({'limit': None, 'offset': 0, 'total': 1})

    assert_success(
        client.retrieve(site_uri, name='Test Site'),
        expected
    )


def test_update(client):
    site_uri = reverse('site-list')

    site_resp = client.create(site_uri, name='Test Site')
    site = TestSite(site_resp.json()['data']['site'])
    site_obj_uri = site.detail_uri()

    # Update the Site name
    params = {'name': 'Test Site 2'}
    expected = site_resp.json()['data']
    expected['site'].update(params)

    assert_success(
        client.update(site_obj_uri, **params),
        expected
    )

    # Now add a description and change the name back
    params = {'name': 'Test Site', 'description': 'A description.'}
    expected['site'].update(params)

    assert_success(
        client.update(site_obj_uri, **params),
        expected
    )

    # Try to update with a required field missing (name)
    assert_error(
        client.update(site_obj_uri, description='Only description.'),
        status.HTTP_400_BAD_REQUEST
    )

    # Or an empty payload
    assert_error(client.update(site_obj_uri), status.HTTP_400_BAD_REQUEST)


def test_deletion(client):
    site_uri = reverse('site-list')
    site_resp = client.create(site_uri, name='Test Site')
    site = TestSite(site_resp.json()['data']['site'])

    # URIs
    site_obj_uri = site.detail_uri()
    attr_uri = site.list_uri('attribute')
    net_uri = site.list_uri('network')

    # Create a network Attribute
    attr_resp = client.create(attr_uri, resource_name='Network', name='attr1')
    attr = attr_resp.json()['data']['attribute']
    attr_obj_uri = site.detail_uri('attribute', id=attr['id'])

    # Create a Network using that Attribute
    net_resp = client.create(net_uri, cidr='10.0.0.0/24', attributes={'attr1': 'foo'})
    net = net_resp.json()['data']['network']
    net_obj_uri = site.detail_uri('network', id=net['id'])

    # Don't allow delete when there's an attached network/attribute
    assert_error(client.delete(site_obj_uri), status.HTTP_409_CONFLICT)
    client.delete(net_obj_uri)

    # Stile Don't allow delete when there's an attached attribute
    assert_error(client.delete(site_obj_uri), status.HTTP_409_CONFLICT)
    client.delete(attr_obj_uri)

    # Finally delete.
    assert_deleted(client.delete(site_obj_uri))



from django.core.urlresolvers import reverse
import json
import logging
import pytest
from pytest_django.fixtures import live_server, django_user_model
import requests

from .util import Client, TestSite


log = logging.getLogger(__name__)


@pytest.fixture
def user(django_user_model):
    """Create and return a non-admin user."""
    user = django_user_model.objects.create(email='user@localhost')
    return user


@pytest.fixture
def site(live_server):
    client = Client(live_server)
    site_uri = reverse('site-list')  # /api/sites/
    resp = client.create(site_uri, name='Test Site')

    site = TestSite(resp.json()['data']['site'])
    return site


@pytest.fixture
def client(live_server):
    return Client(live_server)



"""
Django settings for nsot project using Django 1.8.
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os

# Path where the code is found. (aka project root)
BASE_DIR = os.path.normpath(os.path.join(os.path.dirname(__file__), os.pardir))

# Path where the config file is found.
CONF_ROOT = os.path.abspath(os.path.dirname(__file__))

# A boolean that turns on/off debug mode. Never deploy a site into production
# with DEBUG turned on.
# Default: False
DEBUG = False

#################
# Core Settings #
#################

# A tuple of strings designating all applications that are enabled in this
# Django installation.
INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django_extensions',
    'django_filters',
    'polymorphic',
    'rest_framework',
    'rest_framework_swagger',
    'custom_user',
    'nsot',
)

# The model to use to represent a User.
AUTH_USER_MODEL = 'nsot.User'

# A tuple of authentication backend classes (as strings) to use when attempting
# to authenticate a user.
# https://docs.djangoproject.com/en/1.8/topics/auth/customizing/#authentication-backends
AUTHENTICATION_BACKENDS = (
    'nsot.middleware.auth.EmailHeaderBackend',
    'django.contrib.auth.backends.ModelBackend',
)

# A tuple of middleware classes to use.
# https://docs.djangoproject.com/en/1.8/topics/http/middleware/
MIDDLEWARE_CLASSES = (
    'nsot.middleware.request_logging.LoggingMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'nsot.middleware.auth.EmailHeaderMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.middleware.security.SecurityMiddleware',
)

# A string representing the full Python import path to your root URLconf.
ROOT_URLCONF = 'nsot.conf.urls'

# The full Python path of the WSGI application object that Django's built-in
# servers (e.g. runserver) will use.
WSGI_APPLICATION = 'nsot.wsgi.application'

# When set to True, if the request URL does not match any of the patterns in the
# URLconf and it doesn't end in a slash, an HTTP redirect is issued to the same
# URL with a slash appended. Note that the redirect may cause any data submitted
# in a POST request to be lost.
# Default: True
APPEND_SLASH = True

# Template loaders. The NSoT web UI is written in Angular.js using Jinja2
# templates.
TEMPLATES = [
    {
        "BACKEND": "django_jinja.backend.Jinja2",
        "APP_DIRS": False,
        "OPTIONS": {
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.csrf",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
            ],
            "extensions": [
                "jinja2.ext.do",
                "jinja2.ext.loopcontrols",
                "jinja2.ext.with_",
                "jinja2.ext.i18n",
                "jinja2.ext.autoescape",
                "django_jinja.builtins.extensions.CsrfExtension",
                "django_jinja.builtins.extensions.CacheExtension",
                "django_jinja.builtins.extensions.TimezoneExtension",
                "django_jinja.builtins.extensions.UrlsExtension",
                "django_jinja.builtins.extensions.StaticFilesExtension",
                "django_jinja.builtins.extensions.DjangoFiltersExtension",
            ],
        }
    },
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(BASE_DIR, 'templates').replace('\\','/'),
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                "django.template.context_processors.csrf",
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

##################
# REST Framework #
##################

# Settings for Django REST Framework (DRF)
REST_FRAMEWORK = {
    'DEFAULT_FILTER_BACKENDS': ('rest_framework.filters.DjangoFilterBackend',),
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.IsAdminUser',
    ),
    'DEFAULT_PAGINATION_CLASS': 'nsot.api.pagination.CustomPagination',
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'nsot.api.auth.AuthTokenAuthentication',
        'nsot.api.auth.EmailHeaderAuthentication',
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.SessionAuthentication',
    ),
    'EXCEPTION_HANDLER': 'nsot.exc.custom_exception_handler',
    'TEST_REQUEST_DEFAULT_FORMAT': 'json',
    'PAGE_SIZE': None,
}

############
# Database #
############
# https://docs.djangoproject.com/en/1.8/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'nsot.sqlite3'),
    }
}

###############
# Application #
###############

# The address on which the application will listen.
# Default: localhost
NSOT_HOST = '0.0.0.0'

# The port on which the application will be accessed.
# Default: 8990
NSOT_PORT = 8990

# If True, serve static files directly from the app.
# Default: True
SERVE_STATIC_FILES = True

# Header to check for Authenticated Email. This is intended for use behind an
# authenticating reverse proxy.
USER_AUTH_HEADER = 'X-NSoT-Email'

# The age, in seconds, until an AuthToken granted by the API will expire.
# Default: 600
AUTH_TOKEN_EXPIRY = 600  # 10 minutes

# A list of strings representing the host/domain names that this Django site can
# serve. This is a security measure to prevent an attacker from poisoning caches
# and triggering password reset emails with links to malicious hosts by
# submitting requests with a fake HTTP Host header, which is possible even under
# many seemingly-safe web server configurations.
# https://docs.djangoproject.com/en/1.8/ref/settings/#allowed-hosts
ALLOWED_HOSTS = ['*']

# The name of the cookie to use for the CSRF authentication token.
# Default: 'csrftoken'
CSRF_COOKIE_NAME = '_xsrf'

# Coming in Django 1.9
# The name of the request header used for CSRF authentication.
# https://github.com/django/django/commit/668d53c
# Default: 'X-CSRFToken'
# CSRF_HEADER_NAME = 'X-XSRFToken'

################
# Static files #
################

# (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.8/howto/static-files/

# URL to use when referring to static files located in STATIC_ROOT.
# Default: '/static/'
STATIC_URL = '/static/'

# This setting defines the additional locations the staticfiles app will
# traverse if the FileSystemFinder finder is enabled, e.g. if you use the
# collectstatic or findstatic management command or use the static file serving
# view.
# Default: $BASE_DIR/static
STATICFILES_DIRS = (
    os.path.join(BASE_DIR, 'static'),
)

# The absolute path to the directory where collectstatic will collect static
# files for deployment.
# Default: $BASE_DIR/static'
STATIC_ROOT = os.path.realpath(os.path.join(BASE_DIR, 'static'))

###########
# Swagger #
###########

from nsot.version import __version__
NSOT_VERSION = __version__

SWAGGER_SETTINGS = {
    'exclude_namespaces': ['.*', 'attribute_types'],
    #'api_version': __version__,
}


###########
# Logging #
###########
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format' : "%(asctime)s %(levelname)s [%(name)s:%(lineno)s] %(message)s",
            'datefmt' : "%d/%b/%Y %H:%M:%S"
        },
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
    },
    'handlers': {
        'console':{
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        'django.request': {
            'handlers':['console'],
            'propagate': True,
            'level':'ERROR',
        },
        'nsot': {
            'handlers': ['console'],
            'level': 'INFO',
        },
        'rest_framework': {
            'handlers': ['console'],
            'level': 'ERROR',
        },
    }
}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest
# Allow everything in there to access the DB
pytestmark = pytest.mark.django_db

from django.db import IntegrityError
from django.db.models import ProtectedError
from django.core.exceptions import ValidationError as DjangoValidationError
import logging

from nsot import exc, models

from .fixtures import admin_user, user, site, transactional_db


def test_networks_creation_reparenting(site):
    net_8  = models.Network.objects.create(site=site, cidr=u'10.0.0.0/8')
    net_24 = models.Network.objects.create(site=site, cidr=u'10.0.0.0/24')
    net_16 = models.Network.objects.create(site=site, cidr=u'10.0.0.0/16')
    net_0  = models.Network.objects.create(site=site, cidr=u'0.0.0.0/0')

    # Need to refresh the objects from the db so the updated parent_ids are
    # reflected.
    net_8.refresh_from_db()
    net_24.refresh_from_db()
    net_16.refresh_from_db()
    net_0.refresh_from_db()

    assert net_0.parent_id is None
    assert net_8.parent_id == net_0.id
    assert net_16.parent_id == net_8.id
    assert net_24.parent_id == net_16.id

    assert list(net_0.supernets()) == []
    assert list(
        net_0.subnets().order_by('network_address', 'prefix_length')
    ) == [net_8, net_16, net_24]

    assert list(net_8.supernets()) == [net_0]
    assert list(net_8.subnets()) == [net_16, net_24]

    assert list(net_16.supernets()) == [net_0, net_8]
    assert list(net_16.subnets()) == [net_24]

    assert list(
        net_24.supernets().order_by('network_address', 'prefix_length')
    ) == [net_0, net_8, net_16]
    assert list(net_24.subnets()) == []


def test_seqential_creation(site):
    net_8 = models.Network.objects.create(site=site, cidr=u'10.0.0.0/8')
    net_22_1 = models.Network.objects.create(site=site, cidr=u'10.0.0.0/22')
    net_22_2 = models.Network.objects.create(site=site, cidr=u'10.0.4.0/22')
    net_22_3 = models.Network.objects.create(site=site, cidr=u'10.0.8.0/22')

    net_8.refresh_from_db()
    net_22_1.refresh_from_db()
    net_22_2.refresh_from_db()
    net_22_3.refresh_from_db()

    assert net_8.id == net_22_1.parent_id
    assert net_8.id == net_22_2.parent_id
    assert net_8.id == net_22_3.parent_id


def test_network_create_hostbits_set(site):
    with pytest.raises(exc.ValidationError):
        models.Network.objects.create(site=site, cidr=u'10.0.0.0/0')


def test_network_attributes(site):
    models.Attribute.objects.create(
        site=site,
        resource_name='Network', name='vlan'
    )

    network = models.Network.objects.create(
        site=site, cidr=u'10.0.0.0/8', attributes={'vlan': '34' }
    )

    assert network.get_attributes() == {'vlan': '34'}

    # Verify property successfully zeros out attributes
    network.set_attributes({})
    assert network.get_attributes() == {}

    with pytest.raises(exc.ValidationError):
        network.set_attributes(None)

    with pytest.raises(exc.ValidationError):
        network.set_attributes({0: 'value'})

    with pytest.raises(exc.ValidationError):
        network.set_attributes({'key': 0})

    with pytest.raises(exc.ValidationError):
        network.set_attributes({'made_up': 'value'})


def test_ip_address_no_network(site):
    with pytest.raises(exc.ValidationError):
        models.Network.objects.create(site=site, cidr=u'10.0.0.1/32')

    models.Network.objects.create(site=site, cidr=u'10.0.0.0/8')
    models.Network.objects.create(site=site, cidr=u'10.0.0.1/32')


def test_retrieve_networks(site):
    models.Attribute.objects.create(
        site=site, resource_name='Network', name='test'
    )

    net_8 = models.Network.objects.create(
        site=site, cidr=u'10.0.0.0/8', attributes={'test': 'foo'}
    )
    net_24 = models.Network.objects.create(
        site=site, cidr=u'10.0.0.0/24', attributes={'test': 'bar'}
    )
    ip = models.Network.objects.create(
        site=site, cidr=u'10.0.0.1/32'
    )

    # root=True
    assert list(site.networks.filter(parent_id=None)) == [net_8]


    # include_networks=True, include_ips=Fals
    assert list(site.networks.filter(is_ip=False)) == [net_8, net_24]

    # include_networks=False, include_ips=False
    assert list(site.networks.none()) == []

    # include_networks=True, include_ips=True
    assert list(site.networks.all()) == [net_8, net_24, ip]

    # include_networks=False, include_ips=True
    assert list(site.networks.filter(is_ip=True)) == [ip]

    # There is no use-case for these tests.
    '''
    assert sorted(site.networks(
        subnets_of='10.0.0.0/10'
    )) == sorted([net_24])

    assert sorted(site.networks(
        subnets_of='10.0.0.0/10', include_ips=True
    )) == sorted([net_24, ip])

    assert sorted(site.networks(
        supernets_of='10.0.0.0/10'
    )) == sorted([net_8])

    with pytest.raises(ValueError):
        site.networks(subnets_of='10.0.0.0/10', supernets_of='10.0.0.0/10')
    '''

    with pytest.raises(ValueError):
        assert site.networks.filter(
            attributes__name=None, attributes__value='foo'
        )

    assert list(
        site.networks.filter(attributes__attribute__name='test').order_by('id')
    ) == [net_8, net_24]

    assert list(
        site.networks.filter(
            attributes__attribute__name='test', attributes__value='foo'
        )
    ) == [net_8]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from base64 import b64decode, b64encode
from django.db.backends.sqlite3.base import DatabaseWrapper
from django.db import models
from django.conf import settings
from django.utils.datastructures import DictWrapper
from django.utils.encoding import force_bytes
import ipaddress
import logging
import types


log = logging.getLogger(__name__)


if not hasattr(DatabaseWrapper, 'get_new_connection_is_patched'):
    """
    Monkey-patch SQLite3 driver to handle text as bytes.

    Credit: http://stackoverflow.com/a/28794677/194311
    """
    _get_new_connection = DatabaseWrapper.get_new_connection
    def _get_new_connection_tolerant(self, conn_params):
        conn = _get_new_connection( self, conn_params )
        conn.text_factory = bytes
        return conn

    DatabaseWrapper.get_new_connection = types.MethodType(
        _get_new_connection_tolerant, None, DatabaseWrapper
    )
    DatabaseWrapper.get_new_connection_is_patched = True


class BinaryIPAddressField(models.Field):
    """IP Address field that stores values as varbinary."""
    __metaclass__ = models.SubfieldBase

    def __init__(self, *args, **kwargs):
        super(BinaryIPAddressField, self).__init__(*args, **kwargs)
        self.editable = True

    def db_type(self, connection):
        data = DictWrapper(self.__dict__, connection.ops.quote_name, "qn_")
        return 'varbinary(%(max_length)s)' % data

    def to_python(self, value):
        """DB -> Python."""
        if not value:
            return value

        try:
            return ipaddress.ip_address(value).exploded
        except ValueError:
            return ipaddress.ip_address(bytes(value)).exploded

    def get_db_prep_value(self, value, connection, prepared=False):
        """Python -> DB."""
        return ipaddress.ip_address(value).packed



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest

# Allow everything in there to access the DB
pytestmark = pytest.mark.django_db

import copy
from django.core.urlresolvers import reverse
import json
import logging
from rest_framework import status

from .fixtures import live_server, client, user, site
from .util import (
    assert_created, assert_error, assert_success, assert_deleted, load_json,
    Client, load, filter_devices
)


log = logging.getLogger(__name__)


def test_creation(live_server, user, site):
    admin_client = Client(live_server, user='admin')
    user_client = Client(live_server, user='user')

    # URIs
    site_uri = site.list_uri()
    attr_uri = site.list_uri('attribute')
    dev_uri = site.list_uri('device')

    admin_client.create(attr_uri, resource_name='Device', name='attr1')

    # Invalid permissions
    assert_error(
        user_client.create(
            dev_uri, hostname='device1', attributes={'attr1': 'foo'}
        ),
        status.HTTP_403_FORBIDDEN
    )

    # Missing required field (hostname)
    assert_error(
        admin_client.create(dev_uri, attributes={'attr1': 'foo'}),
        status.HTTP_400_BAD_REQUEST
    )

    # Null hostname
    assert_error(
        admin_client.create(dev_uri, hostname=None),
        status.HTTP_400_BAD_REQUEST
    )

    # Verify successful creation
    dev_resp = admin_client.create(
        dev_uri, hostname='device1', attributes={'attr1': 'foo'}
    )
    dev = dev_resp.json()['data']['device']
    dev_obj_uri = site.detail_uri('device', id=dev['id'])

    assert_created(dev_resp, dev_obj_uri)

    # Verify successful get of all Devices
    expected = dev_resp.json()['data']
    expected['devices'] = [expected.pop('device')]
    expected.update({'limit': None, 'offset': 0, 'total': 1})

    assert_success(admin_client.get(dev_uri), expected)

    # Verify successful get of single Device
    assert_success(admin_client.get(dev_obj_uri), {'device': dev})


def test_collection_creation(site, client):
    # URIs
    dev_uri = site.list_uri('device')

    # Successfully create a collection of Devices
    collection = [
        {'hostname': 'device1'},
        {'hostname': 'device2'},
        {'hostname': 'device3'},
    ]
    collection_response = client.post(
        dev_uri,
        data=json.dumps(collection)
    )
    assert_created(collection_response, None)

    # Successfully get all created Devices
    output = collection_response.json()
    output['data'].update({
        'limit': None, 'offset': 0, 'total': len(collection)
    })

    assert_success(client.get(dev_uri), output['data'])


def test_filters(site, client):
    """Test hostname/attribute filters for Devices."""

    # URIs
    attr_uri = site.list_uri('attribute')
    dev_uri = site.list_uri('device')

    # Pre-load the Attributes
    client.post(attr_uri, data=load('attributes.json'))

    # Populate the Device objects
    dev_resp = client.post(dev_uri, data=load('devices.json'))
    devices_out = dev_resp.json()['data']
    devices = devices_out['devices']

    # Test lookup by hostname
    expected = copy.deepcopy(devices_out)
    wanted = ['foo-bar3']
    expected['devices'] = filter_devices(devices, wanted)
    expected.update({'limit': None, 'offset': 0, 'total': len(wanted)})
    assert_success(
        client.retrieve(dev_uri, hostname='foo-bar3'),
        expected
    )

    # Test lookup by attributes
    wanted = ['foo-bar2', 'foo-bar3']
    expected['devices'] = filter_devices(devices, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(dev_uri, attributes='foo=baz'),
        expected
    )

    # Test lookup with multiple attributes
    wanted = ['foo-bar3']
    expected['devices'] = filter_devices(devices, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(dev_uri, attributes=['foo=baz', 'cluster=lax']),
        expected
    )


def test_set_queries(client, site):
    """Test set queries for Devices."""

    # URIs
    attr_uri = site.list_uri('attribute')
    dev_uri = site.list_uri('device')
    query_uri = site.query_uri('device')

    # Pre-load the attributes
    client.post(attr_uri, data=load('attributes.json'))

    # Populate the device objects.
    dev_resp = client.post(dev_uri, data=load('devices.json'))
    devices_out = dev_resp.json()['data']
    devices = devices_out['devices']

    # INTERSECTION: foo=bar
    expected = copy.deepcopy(devices_out)
    wanted = ['foo-bar1', 'foo-bar4']
    expected['devices'] = filter_devices(devices, wanted)
    expected.update({'limit': None, 'offset': 0, 'total': len(wanted)})
    assert_success(
        client.retrieve(query_uri, query='foo=bar'),
        expected
    )

    # INTERSECTION: foo=bar owner=jathan
    wanted = ['foo-bar1']
    expected['devices'] = filter_devices(devices, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(query_uri, query='foo=bar owner=jathan'),
        expected
    )

    # DIFFERENCE: -owner=gary
    wanted = ['foo-bar1', 'foo-bar3']
    expected['devices'] = filter_devices(devices, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(query_uri, query='-owner=gary'),
        expected
    )

    # UNION: cluster +foo=baz
    wanted = ['foo-bar1', 'foo-bar2', 'foo-bar3']
    expected['devices'] = filter_devices(devices, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(query_uri, query='cluster +foo=baz'),
        expected
    )


def test_update(live_server, user, site):
    admin_client = Client(live_server, user='admin')
    user_client = Client(live_server, user='user')

    # URIs
    site_uri = site.list_uri()
    attr_uri = site.list_uri('attribute')
    dev_uri = site.list_uri('device')

    admin_client.create(attr_uri, resource_name='Device', name='attr1')
    dev_resp = admin_client.create(
        dev_uri, hostname='device1', attributes={'attr1': 'foo'}
    )

    # Extract the device object from the response payload so we can play with
    # it during update tests.
    device = dev_resp.json()['data']['device']
    dev_obj_uri = site.detail_uri('device', id=device['id'])

    # Empty update should only clear attributes.
    params = {'hostname': 'foo'}
    device.update(params)
    device['attributes'] = {}

    assert_success(
        admin_client.update(dev_obj_uri, **params),
        {'device': device}
    )

    # Now put attributes back and change hostname
    params = {'hostname': 'bar', 'attributes': {'attr1': 'foo'}}
    device.update(params)

    assert_success(
        admin_client.update(dev_obj_uri, **params),
        {'device': device}
    )

    # Invalid permissions
    assert_error(user_client.update(dev_obj_uri), status.HTTP_403_FORBIDDEN)


def test_deletion(site, client):
    dev_uri = site.list_uri('device')
    attr_uri = site.list_uri('attribute')

    client.create(attr_uri, resource_name='Device', name='attr1')

    # Create one Device with an Attribute so that we can confirm is is safely
    # deleted.
    dev1_resp = client.create(
        dev_uri, hostname='device1', attributes={'attr1': 'foo'}
    )
    dev1 = dev1_resp.json()['data']['device']
    dev1_obj_uri = site.detail_uri('device', id=dev1['id'])

    # Device 2
    client.create(dev_uri, hostname='device2')

    # Delete Device 1 w/ Attribute
    assert_deleted(client.delete(dev1_obj_uri))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest
# Allow everything in there to access the DB
pytestmark = pytest.mark.django_db

from django.db import IntegrityError
from django.db.models import ProtectedError
from django.core.exceptions import ValidationError as DjangoValidationError
import logging

from nsot import exc, models

from .fixtures import admin_user, user, site, transactional_db


def test_creation(site):
    attr = models.Attribute.objects.create(
        resource_name='Network',
        site=site, name='test_attribute'
    )

    attributes = models.Attribute.objects.all()

    assert attributes.count() == 1
    assert attributes[0].id == attr.id
    assert attributes[0].site_id == site.id
    assert attributes[0].name == attr.name
    assert attributes[0].required == attr.required


def test_conflict(site):
    models.Attribute.objects.create(
        resource_name='Network',
        site=site, name='test_attribute'
    )

    with pytest.raises(DjangoValidationError):
        models.Attribute.objects.create(
            resource_name='Network',
            site=site, name='test_attribute'
        )

    models.Attribute.objects.create(
        resource_name='Network',
        site=site, name='test_attribute_2'
    )


def test_validation(site, transactional_db):
    with pytest.raises(exc.ValidationError):
        models.Attribute.objects.create(
            resource_name='Network',
            site=site, name=None,
        )

    with pytest.raises(exc.ValidationError):
        models.Attribute.objects.create(
            resource_name='Network',
            site=site, name='',
        )

    attribute = models.Attribute.objects.create(
        resource_name='Network',
        site=site, name='test_attribute'
    )

    with pytest.raises(exc.ValidationError):
        attribute.name = ''
        attribute.save()

    with pytest.raises(exc.ValidationError):
        attribute.name = None
        attribute.save()

    attribute.name = 'test_attribute_new'
    attribute.save()


def test_deletion(site):
    attribute = models.Attribute.objects.create(
        resource_name='Network',
        site=site, name='test_attribute'
    )

    network = models.Network.objects.create(
        cidr='10.0.0.0/8', site=site, attributes={'test_attribute': 'foo'}
    )

    with pytest.raises(ProtectedError):
        attribute.delete()

    network.delete()
    attribute.delete()


def test_required(site):
    attribute_1 = models.Attribute.objects.create(
        resource_name='Network',
        site=site, name='required_1', required=True
    )
    attribute_2 = models.Attribute.objects.create(
        resource_name='Network',
        site=site, name='required_2', required=True
    )

    with pytest.raises(exc.ValidationError):
        network = models.Network.objects.create(
            cidr='10.0.0.0/8', site=site, attributes={}
        )

    with pytest.raises(exc.ValidationError):
        network = models.Network.objects.create(
            site=site, cidr='10.0.0.0/8', attributes={'required_1': 'foo'}
        )

    network = models.Network.objects.create(
        cidr=u'10.0.0.0/8',
        attributes={'required_1': 'foo', 'required_2': 'bar'},
        site=site,
    )


def test_multi(site):
    multi = models.Attribute.objects.create(
        resource_name='Network', display=True,
        site=site, name='multi', multi=True
    )

    not_multi = models.Attribute.objects.create(
        resource_name='Network',
        site=site, name='not_multi', multi=False
    )

    models.Network.objects.create(site=site, cidr='10.0.0.0/8')

    network = models.Network.objects.create(
        site=site, cidr='10.0.0.1',
        attributes={'multi': ['test', 'testing', 'testtttt']}
    )

    with pytest.raises(exc.ValidationError):
        network = models.Network.objects.create(
            site=site, cidr=u'10.0.0.2',
            attributes={'not_multi': ['test', 'testing', 'testtttt']}
        )


def test_constraints(site):
    default = models.Attribute.objects.create(
        resource_name='Network', site=site, name='default'
    )

    allow_empty = models.Attribute.objects.create(
        resource_name='Network', site=site, name='allow_empty',
        constraints={'allow_empty': True}
    )

    pattern = models.Attribute.objects.create(
        resource_name='Network', site=site, name='pattern',
        constraints={'pattern': '\d\d\d+'}
    )

    valid = models.Attribute.objects.create(
        resource_name='Network', site=site, name='valid',
        constraints={'valid_values': ['foo', 'bar', 'baz']}
    )

    # Test that ValidationError is raised when constraints are not a dict
    with pytest.raises(exc.ValidationError):
        models.Attribute.objects.create(
            resource_name='Network', site=site,
            name='invalid', constraints=['foo', 'bar', 'baz']
        )

    network = models.Network.objects.create(site=site, cidr='10.0.0.0/8')

    with pytest.raises(exc.ValidationError):
        network.set_attributes({'default': ''})

    # Test allow_empty
    network.set_attributes({'allow_empty': ''})

    # Test pattern
    with pytest.raises(exc.ValidationError):
        network.set_attributes({'pattern': ''})

    with pytest.raises(exc.ValidationError):
        network.set_attributes({'pattern': 'foo'})

    with pytest.raises(exc.ValidationError):
        network.set_attributes({'pattern': '10'})

    network.set_attributes({'pattern': '100'})
    network.set_attributes({'pattern': '1000000'})

    # Test valid_values
    with pytest.raises(exc.ValidationError):
        network.set_attributes({'valid': 'hello'})

    network.set_attributes({'valid': 'foo'})



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest
# Allow everything in there to access the DB
pytestmark = pytest.mark.django_db

from django.db import IntegrityError
from django.db.models import ProtectedError
from django.core.exceptions import ValidationError as DjangoValidationError
import logging

from nsot import exc, models

from .fixtures import user, transactional_db


log = logging.getLogger(__name__)


def test_site_creation():
    site = models.Site.objects.create(
        name='Test Site',
        description='This is a Test Site.'
    )
    sites = models.Site.objects.all()

    assert sites.count() == 1
    assert sites[0].id == site.id
    assert sites[0].name == site.name
    assert sites[0].description == site.description


def test_site_conflict(transactional_db):
    models.Site.objects.create(
        name='Test Site',
        description='This is a Test Site.'
    )

    with pytest.raises(DjangoValidationError):
        models.Site.objects.create(
            name='Test Site',
            description='This is a Test Site.'
        )

    models.Site.objects.create(
        name='Test Site 2',
        description='This is a Test Site.'
    )


def test_site_validation(transactional_db):
    with pytest.raises(exc.ValidationError):
        models.Site.objects.create(
            name=None,
            description='This is a Test Site.'
        )

    with pytest.raises(exc.ValidationError):
        models.Site.objects.create(
            name='',
            description='This is a Test Site.'
        )

    site = models.Site.objects.create(
        name='Test Site',
        description='This is a Test Site.'
    )

    with pytest.raises(exc.ValidationError):
        site.name = ''
        site.save()

    with pytest.raises(exc.ValidationError):
        site.name = None
        site.save()

    site.name = 'Test Site New'
    site.save()



from __future__ import unicode_literals

from collections import OrderedDict
from django import http
import logging
from rest_framework.exceptions import ValidationError
from rest_framework.exceptions import APIException
from rest_framework.views import exception_handler


log = logging.getLogger(__name__)


def custom_exception_handler(exc, context):
    """Always handle errors all pretty-like."""
    # Call REST framework's default exception handler first to get the standard
    # error response.
    response = exception_handler(exc, context)

    # Now add the HTTP status code and message to the response.
    # We want an error response to look like this:
    # {
    #     "status": "error",
    #     "error": {
    #         "message": "Endpoint not found",
    #         "code": 404
    #     }
    # }
    log.debug('custom_exception_handler: exc = %r', exc)
    log.debug('custom_exception_handler: context = %r', context)
    if response is not None:
        orig_data = response.data
        log.debug('custom_exception_handler: orig_data = %r', orig_data)

        try:
            message = orig_data['detail']
            if message == 'Not found.':
                message = 'Endpoint not found.'
        except KeyError:
            message = orig_data
        except TypeError:
            message = orig_data[0]

        data = OrderedDict([
            ('status', 'error'),
            ('error', {
                'message': message,
                'code': response.status_code,
            }),
        ])
        response.data = data

    request = context['request']
    log.debug('custom_exception_handler: request = %r', request)
    log.debug('custom_exception_handler: dir(request) = %r', dir(request))
    log.debug('custom_exception_handler: request.data = %r', request.data)

    return response


class Error(APIException):
    """ Baseclass for NSoT Exceptions."""


class ModelError(Error):
    """Base class for NSoT Model Exceptions."""


class BaseHttpError(Error):
    pass


class BadRequest(BaseHttpError):
    status_code = 400


class Unauthorized(BaseHttpError):
    status_code = 401


class Forbidden(BaseHttpError):
    status_code = 403


class NotFound(BaseHttpError):
    status_code = 404
    default_detail = 'Endpoint not found.'


class Conflict(BaseHttpError):
    status_code = 409



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest

# Allow everything in there to access the DB
pytestmark = pytest.mark.django_db

import copy
from django.core.urlresolvers import reverse
import json
import logging
from rest_framework import status


from .fixtures import live_server, client, user, site
from .util import (
    assert_created, assert_error, assert_success, assert_deleted, load_json,
    Client, load
)


log = logging.getLogger(__name__)


def test_creation(client, site):

    attr_uri = site.list_uri('attribute')

    # Test invalid attribute name
    assert_error(
        client.create(
            attr_uri,
            resource_name='Network', name='invalid attr1'
        ),
        status.HTTP_400_BAD_REQUEST
    )

    # Successfully create an Attribute
    attr_resp = client.create(attr_uri, resource_name='Network', name='attr1')
    attr = attr_resp.json()['data']['attribute']
    attr_obj_uri = site.detail_uri('attribute', id=attr['id'])
    assert_created(attr_resp, attr_obj_uri)

    # Successfully get all Network Attributes
    expected = attr_resp.json()['data']
    expected['attributes'] = [expected.pop('attribute')]
    expected.update({'limit': None, 'offset': 0, 'total': 1})

    assert_success(client.get(attr_uri), expected)

    # Successfully get a single Network Attribute
    assert_success(client.get(attr_obj_uri), attr_resp.json()['data'])


def test_collection_creation(client, site):

    attr_uri = site.list_uri('attribute')

    # Successfully create a collection of Attributes
    collection = [
        {'name': 'attr1', 'resource_name': 'Network'},
        {'name': 'attr2', 'resource_name': 'Network'},
        {'name': 'attr3', 'resource_name': 'Network'},
    ]
    collection_response = client.post(
        attr_uri,
        data=json.dumps(collection)
    )
    assert_created(collection_response, None)

    # Successfully get all created Attributes
    output = collection_response.json()
    output['data'].update({
        'limit': None, 'offset': 0, 'total': len(collection)
    })

    assert_success(
        client.get(attr_uri),
        output['data'],
    )


def test_update(client, site):

    attr_uri = site.list_uri('attribute')

    attr_resp = client.create(attr_uri, resource_name='Network', name='attr1')
    attr = attr_resp.json()['data']['attribute']
    attr_obj_uri = site.detail_uri('attribute', id=attr['id'])

    # Update the description
    params = {'description': 'Attribute 1'}
    attr1 = copy.deepcopy(attr)
    attr1.update(params)

    assert_success(
        client.update(attr_obj_uri, **params),
        {'attribute': attr1}
    )

    # Update the required flag; which should also set display=True
    params = {'required': True}
    attr2 = copy.deepcopy(attr1)
    attr2.update(params)
    attr2['display'] = True

    assert_success(
        client.update(attr_obj_uri, **params),
        {'attribute': attr2}
    )

    # Reset the object back to it's initial state!
    assert_success(
        client.update(attr_obj_uri, **attr),
        {'attribute': attr}
    )


def test_deletion(client, site):

    attr_uri = site.list_uri('attribute')
    net_uri = site.list_uri('network')

    attr_resp = client.create(attr_uri, resource_name='Network', name='attr1')
    attr = attr_resp.json()['data']['attribute']
    attr_obj_uri = site.detail_uri('attribute', id=attr['id'])

    # Create a Network with an attribute
    net_resp = client.create(
        net_uri, cidr='10.0.0.0/24', attributes={'attr1': 'foo'}
    )
    net = net_resp.json()['data']['network']
    net_obj_uri = site.detail_uri('network', id=net['id'])

    # Don't allow delete when there's an attached network
    assert_error(
        client.delete(attr_obj_uri),
        status.HTTP_409_CONFLICT
    )

    # Now delete the Network
    client.delete(net_obj_uri)

    # And safely delete the Attribute
    assert_deleted(client.delete(attr_obj_uri))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf.urls import url, include
from django.conf import settings
from rest_framework import routers
from rest_framework_nested import routers as nested_routers

from . import views


# Register /sites as a top-level resource
router = routers.DefaultRouter(trailing_slash=settings.APPEND_SLASH)
router.register(r'sites', views.SiteViewSet)

# Nested router for resources under /sites
sites_router = nested_routers.NestedSimpleRouter(
    router, r'sites', lookup='site', trailing_slash=settings.APPEND_SLASH
)

# Resources that are nested under /sites
sites_router.register(r'attributes', views.AttributeViewSet)
sites_router.register(r'changes', views.ChangeViewSet)
sites_router.register(r'devices', views.DeviceViewSet)
sites_router.register(r'networks', views.NetworkViewSet)

# Resources pinned to API index at /
router.register(r'attributes', views.AttributeViewSet)
router.register(r'changes', views.ChangeViewSet)
router.register(r'devices', views.DeviceViewSet)
router.register(r'networks', views.NetworkViewSet)
router.register(r'users', views.UserViewSet)
router.register(r'values', views.ValueViewSet)

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browsable API.
urlpatterns = [
    # API routes
    url(r'^', include(router.urls)),
    url(r'^', include(sites_router.urls)),

    # Browsable API auth login
    url(r'^api-auth/', include('rest_framework.urls',
        namespace='rest_framework')),

    # API auth_token login/verify (email/secret_key)
    url(r'^authenticate/', views.AuthTokenLoginView.as_view(),
        name='authenticate'),
    url(r'^verify_token/', views.AuthTokenVerifyView.as_view(),
        name='verify_token'),
]



from __future__ import absolute_import, print_function

from django.core.management import call_command
from django.core.management.base import BaseCommand

from optparse import make_option


class Command(BaseCommand):
    help = 'Performs any pending database migrations and upgrades'

    option_list = BaseCommand.option_list + (
        make_option('--noinput',
            action='store_true',
            dest='noinput',
            default=False,
            help='Tells Django to NOT prompt the user for input of any kind.',
        ),
    )

    def handle(self, **options):
        call_command(
            'syncdb',
            migrate=True,
            interactive=(not options['noinput']),
            traceback=options['traceback'],
            verbosity=options['verbosity'],
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest

# Allow everything in there to access the DB
pytestmark = pytest.mark.django_db

import copy
from django.core.urlresolvers import reverse
import json
import logging
from rest_framework import status

from .fixtures import live_server, client, user, site
from .util import (
    assert_created, assert_error, assert_success, assert_deleted, load_json,
    Client, load, filter_networks
)


log = logging.getLogger(__name__)


def test_creation(live_server, user, site):
    admin_client = Client(live_server, 'admin')
    user_client = Client(live_server, 'user')

    # URIs
    site_uri = site.list_uri()
    attr_uri = site.list_uri('attribute')
    net_uri = site.list_uri('network')

    admin_client.create(attr_uri, resource_name='Network', name='attr1')

    # Invalid permissions
    assert_error(
        user_client.create(
            net_uri, cidr='10.0.0.0/24', attributes={'attr1': 'foo'}
        ),
        status.HTTP_403_FORBIDDEN
    )

    # Missing required field (cidr)
    assert_error(
        admin_client.create(net_uri, attributes={'attr1': 'foo'}),
        status.HTTP_400_BAD_REQUEST
    )

    # Null cidr
    assert_error(
        admin_client.create(net_uri, cidr=None),
        status.HTTP_400_BAD_REQUEST
    )

    # Verify Successful Creation
    net_resp = admin_client.create(
        net_uri, cidr='10.0.0.0/24', attributes={'attr1': 'foo'}
    )
    net = net_resp.json()['data']['network']
    net_obj_uri = site.detail_uri('network', id=net['id'])

    assert_created(net_resp, net_obj_uri)

    # Verify Successful get of all Networks
    expected = net_resp.json()['data']
    expected['networks'] = [expected.pop('network')]
    expected.update({'limit': None, 'offset': 0, 'total': 1})

    assert_success(admin_client.get(net_uri), expected)

    # Verify Successful get of single Network
    assert_success(admin_client.get(net_obj_uri), {'network': net})


def test_collection_creation(site, client):
    # URIs
    net_uri = site.list_uri('network')

    # Successfully create a collection of Networks
    collection = [
        {'cidr': '1.1.1.0/24'},
        {'cidr': '2.2.2.0/24'},
        {'cidr': '3.3.3.0/24'},
    ]
    collection_response = client.post(
        net_uri,
        data=json.dumps(collection)
    )
    assert_created(collection_response, None)

    # Successfully get all created Networks
    output = collection_response.json()
    output['data'].update({
        'limit': None, 'offset': 0, 'total': len(collection)
    })

    assert_success(client.get(net_uri), output['data'])


def test_filters(site, client):
    """Test cidr/address/prefix/attribute filters for Networks."""

    # URIs
    attr_uri = site.list_uri('attribute')
    net_uri = site.list_uri('network')

    # Pre-load the Attributes
    client.post(attr_uri, data=load('attributes.json'))

    # Populate the network objects and retreive them for testing.
    client.post(net_uri, data=load('networks.json'))
    net_resp = client.retrieve(net_uri, include_ips=True)
    net_out = net_resp.json()['data']
    networks = net_out['networks']

    # Test lookup by cidr
    expected = copy.deepcopy(net_out)
    wanted = ['10.0.0.0/8']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'limit': None, 'offset': 0, 'total': len(wanted)})
    assert_success(
        client.retrieve(net_uri, cidr='10.0.0.0/8'),
        expected
    )

    # Test lookup by attributes
    wanted = ['192.168.0.0/16', '172.16.0.0/12']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(net_uri, attributes='foo=baz'),
        expected
    )

    # Test lookup with multiple attributes
    wanted = ['172.16.0.0/12']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(net_uri, attributes=['foo=baz', 'cluster=lax']),
        expected
    )

    # Test lookup by network_address
    wanted = ['169.254.0.0/16']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(net_uri, network_address='169.254.0.0'),
        expected
    )

    # Test lookup by prefix_length
    wanted = ['192.168.0.0/16', '169.254.0.0/16']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(net_uri, prefix_length=16),
        expected
    )


    # Test lookup by network_address + prefix_length
    wanted = ['10.0.0.0/8']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(net_uri, network_address='10.0.0.0', prefix_length=8),
        expected
    )


def test_set_queries(client, site):
    """Test set queries for Networks."""

    # URIs
    attr_uri = site.list_uri('attribute')
    net_uri = site.list_uri('network')
    query_uri = site.query_uri('network')

    # Pre-load the attributes
    client.post(attr_uri, data=load('attributes.json'))

    # Populate the network objects and retreive them for testing.
    client.post(net_uri, data=load('networks.json'))
    net_resp = client.retrieve(net_uri, include_ips=True)
    net_out = net_resp.json()['data']
    networks = net_out['networks']

    # INTERSECTION: foo=bar
    expected = copy.deepcopy(net_out)
    wanted = ['10.0.0.0/8', '169.254.0.0/16']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'limit': None, 'offset': 0, 'total': len(wanted)})
    assert_success(
        client.retrieve(query_uri, query='foo=bar'),
        expected
    )

    # INTERSECTION: foo=bar owner=jathan
    wanted = ['169.254.0.0/16']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(query_uri, query='foo=bar owner=jathan'),
        expected
    )

    # DIFFERENCE: -owner=gary
    wanted = ['192.168.0.0/16', '169.254.0.0/16', '192.168.0.0/24',
              '192.168.0.0/25']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(query_uri, query='-owner=gary'),
        expected
    )

    # UNION: cluster +foo=baz
    wanted = ['192.168.0.0/16', '172.16.0.0/12']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(query_uri, query='cluster +foo=baz'),
        expected
    )

    # Test that include_ips=True actually does that.
    wanted = ['192.168.0.1/32']
    expected['networks'] = filter_networks(networks, wanted)
    expected.update({'total': len(wanted)})
    assert_success(
        client.retrieve(query_uri, query='vlan=300', include_ips=True),
        expected
    )


def test_update(live_server, user, site):
    admin_client = Client(live_server, 'admin')
    user_client = Client(live_server, 'user')

    # URIs
    site_uri = site.list_uri()
    attr_uri = site.list_uri('attribute')
    net_uri = site.list_uri('network')

    admin_client.create(attr_uri, resource_name='Network', name='attr1')
    net_resp = admin_client.create(
        net_uri, cidr='10.0.0.0/24', attributes={'attr1': 'foo'}
    )

    # Extract the Network object so that we can play w/ it during update tests.
    net = net_resp.json()['data']['network']
    net_obj_uri = site.detail_uri('network', id=net['id'])

    # Empty Update should only clear attributes.
    params = {'attributes': {}}
    net.update(params)

    assert_success(
        admin_client.update(net_obj_uri, **params),
        {'network': net}
    )

    # Now put attributes back
    params['attributes'] = {'attr1': 'foo'}
    net.update(params)

    assert_success(
        admin_client.update(net_obj_uri, **params),
        {'network': net}
    )

    # Invalid permissions
    assert_error(user_client.update(net_obj_uri), status.HTTP_403_FORBIDDEN)


def test_deletion(site, client):
    net_uri = site.list_uri('network')
    attr_uri = site.list_uri('attribute')

    net1_resp = client.create(net_uri, cidr='10.0.0.0/24')
    net1 = net1_resp.json()['data']['network']
    net1_obj_uri = site.detail_uri('network', id=net1['id'])

    net2_resp = client.create(net_uri, cidr='10.0.0.1/32')
    net2 = net2_resp.json()['data']['network']
    net2_obj_uri = site.detail_uri('network', id=net2['id'])

    # Don't allow delete when there's an attached subnet/ip
    assert_error(client.delete(net1_obj_uri), status.HTTP_409_CONFLICT)

    # Delete the child Network
    client.delete(net2_obj_uri)

    # And safely delete the parent Network
    assert_deleted(client.delete(net1_obj_uri))



"""
Configuration file for the unit tests via py.test.
"""

import betamax
import os


CASSETTE_DIR = 'tests/api_tests/cassettes'


# Tell Betamax where to save the cassette fixtures.
with betamax.Betamax.configure() as config:

    if not os.path.exists(CASSETTE_DIR):
        os.makedirs(CASSETTE_DIR)

    config.cassette_library_dir = 'tests/api_tests/cassettes'
    config.default_cassette_options['record_mode'] = 'all'



from django.db import transaction
import ipaddress
import pytest
import time

from nsot.api import exc, models

from model_tests.fixtures import site


@pytest.mark.django_db
def test_create_1024(site):

    address = u'10.0.0.0/20'
    models.Network.objects.create(site=site, cidr=address)
    models.Attribute.objects.create(
        site=site, resource_name='Network', name='aaaa'
    )

    start = time.time()
    network = ipaddress.ip_network(address)

    with transaction.atomic():
        for ip in network.subnets(new_prefix=30):
            models.Network.objects.create(
                site=site, cidr=ip.exploded, attributes={'aaaa': 'value'}
            )

    print 'Finished in {} seconds.'.format(time.time() - start)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest
# Allow everything in there to access the DB
pytestmark = pytest.mark.django_db

from django.db import IntegrityError
from django.db.models import ProtectedError
from django.core.exceptions import ValidationError as DjangoValidationError
import logging

from nsot import exc, models

from .fixtures import admin_user, user, site, transactional_db


def test_device_attributes(site):
    models.Attribute.objects.create(
        site=site,
        resource_name='Device', name='owner'
    )

    device = models.Device.objects.create(
        site=site, hostname='foobarhost', attributes={'owner': 'gary'}
    )

    assert device.get_attributes() == {'owner': 'gary'}

    # Verify property successfully zeros out attributes
    device.set_attributes({})
    assert device.get_attributes() == {}

    with pytest.raises(exc.ValidationError):
        device.set_attributes(None)

    with pytest.raises(exc.ValidationError):
        device.set_attributes({0: 'value'})

    with pytest.raises(exc.ValidationError):
        device.set_attributes({'key': 0})

    with pytest.raises(exc.ValidationError):
        device.set_attributes({'made_up': 'value'})


def test_retrieve_device(site):
    models.Attribute.objects.create(
        site=site,
        resource_name='Device', name='test'
    )

    device1 = models.Device.objects.create(
        site=site, hostname='device1',
        attributes={'test': 'foo'}
    )
    device2 = models.Device.objects.create(
        site=site, hostname='device2',
        attributes={'test': 'bar'}
    )
    device3 = models.Device.objects.create(
        site=site, hostname='device3'
    )

    assert list(site.devices.all()) == [device1, device2, device3]

    with pytest.raises(ValueError):
        assert site.devices.filter(attributes__name=None, attributes__value='foo')

    assert list(
        site.devices.filter(
            attributes__attribute__name='test'
        ).order_by('id')
    ) == [device1, device2]

    assert list(
        site.devices.filter(
            attributes__attribute__name='test', attributes__value='foo'
        )
    ) == [device1]



#!/usr/bin/env python

import os

from setuptools import setup, find_packages

execfile('nsot/version.py')

with open('requirements.txt') as requirements:
    required = requirements.read().splitlines()

package_data = {}
def get_package_data(package, base_dir):
    for dirpath, dirnames, filenames in os.walk(base_dir):
        dirpath = dirpath[len(package)+1:]  # Strip package dir
        for filename in filenames:
            package_data.setdefault(package, []).append(os.path.join(dirpath, filename))
        for dirname in dirnames:
            get_package_data(package, dirname)

get_package_data('nsot', 'nsot/static')
get_package_data('nsot', 'nsot/templates')
get_package_data('nsot', 'nsot/migrations')

kwargs = {
    'name': 'nsot',
    'version': str(__version__),
    'packages': find_packages(exclude=['tests']),
    'package_data': package_data,
    'description': 'Network Source of Truth (IP Address Management).',
    'author': 'Gary M. Josack',
    'maintainer': 'Gary M. Josack',
    'author_email': 'gary@dropbox.com',
    'maintainer_email': 'gary@dropbox.com',
    'license': 'Apache',
    'install_requires': required,
    'url': 'https://github.com/dropbox/nsot',
    'entry_points': """
        [console_scripts]
        nsot-server=nsot.util:main
    """,
    'classifiers': [
        'Programming Language :: Python',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
}

setup(**kwargs)



from django.contrib.auth import backends, middleware
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import EmailValidator
import logging

from .. import exc
from ..util import normalize_auth_header


# log = logging.getLogger(__name__)



class EmailHeaderMiddleware(middleware.RemoteUserMiddleware):
    header = normalize_auth_header(settings.USER_AUTH_HEADER)


class EmailHeaderBackend(backends.RemoteUserBackend):
    """Custom backend that validates username is an email."""
    def authenticate(self, remote_user):
        """Override default to return None if username is invalid."""
        if not remote_user:
            return
        username = self.clean_username(remote_user)
        if not username:
            return

        return super(EmailHeaderBackend, self).authenticate(remote_user)

    def clean_username(self, username):
        """Makes sure that the username is a valid email address."""
        validator = EmailValidator()
        try:
            validator(username)  # If invalid, this will raise a ValidationError
        except ValidationError as err:
            # log.debug('Invalid email address: %r', username)
            return None
        else:
            return username

    def configure_user(self, user):
        """Make all new users superusers and staff"""
        user.is_superuser = True
        user.is_staff = True
        user.save()
        # log.debug('Created new user: %s', user)
        return user
